#include <iostream>
#include <chrono>
#include <fstream>

#define INCBIN_STYLE INCBIN_STYLE_SNAKE
#define INCBIN_PREFIX
#include <incbin.h>

#include "mbface_ir.h"
#include "util.h"

using namespace nncase;
using namespace nncase::runtime;
using namespace nncase::runtime::k230;

INCBIN(mbface_ir_features_dataset, "mbface_ir_features_dataset.bin");
INCBIN(mbface_ir_features_test, "mbface_ir_features_test.bin");

MbfaceIr::MbfaceIr(const char *kmodel_file): Model("MbfaceIr", kmodel_file)
{
    runtime_shape_t in_shape { 1, 1, 1280, 800 };
    ai2d_in_tensor_ = host_runtime_tensor::create(datatype_t::dt_uint16, in_shape, hrt::pool_shared).expect("cannot create input tensor");

    runtime_shape_t out_shape { 1, 1, 224, 224 };
    ai2d_out_tensor_ = host_runtime_tensor::create(datatype_t::dt_uint8, out_shape, hrt::pool_shared).expect("cannot create output tensor");

    // ai2d config
    ai2d_datatype_t ai2d_dtype { ai2d_format::NCHW_FMT, ai2d_format::NCHW_FMT, datatype_t::dt_uint8, datatype_t::dt_uint8 };
    ai2d_crop_param_t crop_param { true, 30, 20, 400, 600 };
    ai2d_shift_param_t shift_param { false, 0 };
    ai2d_pad_param_t pad_param { false, { { 0, 0 }, { 0, 0 }, { 0, 0 }, { 60, 60 } }, ai2d_pad_mode::constant, { 255 } };
    ai2d_resize_param_t resize_param { false, ai2d_interp_method::tf_bilinear, ai2d_interp_mode::half_pixel };
    ai2d_affine_param_t affine_param { true, ai2d_interp_method::cv2_bilinear, 0, 0, 127, 1, { 0.5, 0.1, 0.0, 0.1, 0.5, 0.0 } };
    ai2d_builder_.reset(new ai2d_builder(ai2d_in_tensor_, ai2d_out_tensor_, ai2d_dtype, crop_param, shift_param, pad_param, resize_param, affine_param));
    ai2d_builder_->build_schedule();
}

MbfaceIr::~MbfaceIr()
{
}

void MbfaceIr::preprocess()
{
    {
#if PROFILING
        ScopedTiming st(model_name() + " ai2d run");
#endif

        // run ai2d
        ai2d_builder_->invoke().expect("error occurred in ai2d running");
    }

    {
#if PROFILING
        ScopedTiming st(model_name() + " cpu copy");
#endif

        // update kpu in_tensor
        auto ai2d_mapped = std::move(hrt::map(ai2d_out_tensor_, hrt::map_read).unwrap());
        auto ai2d_out_buf = ai2d_mapped.buffer();
        size_t size = ai2d_out_buf.size();
        auto kpu_in_tensor = input_tensor(0);
        auto kpu_mapped = std::move(hrt::map(kpu_in_tensor, hrt::map_write).unwrap());
        auto kpu_mapped_buf = kpu_mapped.buffer();

        // copy 1 channel to 3 channels
        memcpy(reinterpret_cast<char *>(kpu_mapped_buf.data()), ai2d_out_buf.data(), size);
        memcpy(reinterpret_cast<char *>(kpu_mapped_buf.data() + size), ai2d_out_buf.data(), size);
        memcpy(reinterpret_cast<char *>(kpu_mapped_buf.data() + size * 2), ai2d_out_buf.data(), size);
        hrt::sync(kpu_in_tensor, hrt::sync_write_back, true).expect("write back input failed");
    }
}

void MbfaceIr::postprocess()
{
#if PROFILING
    ScopedTiming st(model_name() + " " + __FUNCTION__);
#endif

    auto tensor = output_tensor(0);
    auto mapped_buf = std::move(hrt::map(tensor, hrt::map_read).unwrap());
    float *output_data = reinterpret_cast<float *>(mapped_buf.buffer().data());

    // postprocess
    float score_max = 0;
    uint32_t score_index = 0;
    int MAX_FEATURES_NUM = 100;
    int dl = 173, tl = 1;
    int out_loop_cnts = MAX_FEATURES_NUM / dl;
    int left_loop_cnts = MAX_FEATURES_NUM - out_loop_cnts * dl;
    for (uint32_t tt = 0; tt < tl; tt++)
    {
        for (uint32_t oo = 0; oo < out_loop_cnts; oo++)
        {
            score_index = calulate_score(reinterpret_cast<const float *>(mbface_ir_features_test_data + tt * FL), reinterpret_cast<const float *>(mbface_ir_features_dataset_data), dl, &score_max);
        }
        score_index = calulate_score(reinterpret_cast<const float *>(mbface_ir_features_test_data + tt * FL), reinterpret_cast<const float *>(mbface_ir_features_dataset_data), left_loop_cnts, &score_max);
    }

#if !PROFILING
    std::cout << model_name() << ": score_index: " << score_index << std::endl;
#endif
}