#include <iostream>
#include <chrono>
#include <fstream>
#include "model.h"
#include "util.h"

template <class T>
std::vector<T> read_binary_file(const char *file_name)
{
    std::ifstream ifs(file_name, std::ios::binary);
    ifs.seekg(0, ifs.end);
    size_t len = ifs.tellg();
    std::vector<T> vec(len / sizeof(T), 0);
    ifs.seekg(0, ifs.beg);
    ifs.read(reinterpret_cast<char *>(vec.data()), len);
    ifs.close();
    return std::move(vec);
}

Model::Model(const char *model_name, const char *kmodel_file): model_name_(model_name)
{
    // load kmodel
    kmodel_ = read_binary_file<unsigned char>(kmodel_file);
    interp_.load_model({ (const gsl::byte *)kmodel_.data(), kmodel_.size() }).expect("cannot load kmodel.");

    // create kpu input tensors
    for (size_t i = 0; i < interp_.inputs_size(); i++)
    {
        auto desc = interp_.input_desc(i);
        auto shape = interp_.input_shape(i);
        auto tensor = host_runtime_tensor::create(desc.datatype, shape, hrt::pool_shared).expect("cannot create input tensor");
        interp_.input_tensor(i, tensor).expect("cannot set input tensor");
    }

    // create kpu output tensors
    for (size_t i = 0; i < interp_.outputs_size(); i++)
    {
        auto desc = interp_.output_desc(i);
        auto shape = interp_.output_shape(i);
        auto tensor = host_runtime_tensor::create(desc.datatype, shape, hrt::pool_shared).expect("cannot create output tensor");
        interp_.output_tensor(i, tensor).expect("cannot set output tensor");
    }
}

Model::~Model()
{

}

void Model::run(std::vector<std::string> &input_files)
{
    // read input into ai2d
    auto input_mapped = std::move(hrt::map(ai2d_in_tensor_, hrt::map_write).unwrap());
    read_binary_file(input_files[0].c_str(), reinterpret_cast<char *>(input_mapped.buffer().data()));
    auto ret = input_mapped.unmap();
    hrt::sync(ai2d_in_tensor_, hrt::sync_write_back, true).expect("write back input failed");

    preprocess();
    kpu_run();
    postprocess();
}

std::string Model::model_name() const
{
    return model_name_;
}

void Model::kpu_run()
{
#if PROFILING
    ScopedTiming st(model_name() + " " + __FUNCTION__);
#endif

    interp_.run().expect("error occurred in running model");
}

runtime_tensor Model::input_tensor(size_t idx)
{
    return interp_.input_tensor(idx).expect("cannot get input tensor");
}

void Model::input_tensor(size_t idx, runtime_tensor &tensor)
{
    interp_.input_tensor(idx, tensor).expect("cannot set input tensor");
}

runtime_tensor Model::output_tensor(size_t idx)
{
    return interp_.output_tensor(idx).expect("cannot get output tensor");
}