/* Copyright 2019-2021 Canaan Inc.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
 */
#pragma once
#include "nncase/ir/ops/k230/ir_types.h"
#include <cmath>
#include <nncase/ir/ops/k230/op_utils.h>
#include <nncase/kernels/kernel_utils.h>
#include <nncase/runtime/nnil.h>
#include <nncase/runtime/runtime_op_utility.h>
#include <stdexcept>
#include <type_traits>
#include <xtl/xspan.hpp>
#ifdef __riscv
#include "../riscv/k230_kernels.h"
#endif

using namespace nncase::ir::k230;
namespace nncase::kernels::k230
{
#define UNFOLD_OUTPUT_ASSIGN_4(_output_ptr) \
    auto *o0 = _output_ptr + 0;             \
    auto *o1 = _output_ptr + 1;             \
    auto *o2 = _output_ptr + 2;             \
    auto *o3 = _output_ptr + 3;

#define UNFOLD_ASSIGN_4(_input_ptr, _output_ptr) \
    auto *i0 = _input_ptr + 0;                   \
    auto *i1 = _input_ptr + 1;                   \
    auto *i2 = _input_ptr + 2;                   \
    auto *i3 = _input_ptr + 3;                   \
    UNFOLD_OUTPUT_ASSIGN_4(_output_ptr)

#define UNFOLD_CAST_4()         \
    *o0 = static_cast<TO>(*i0); \
    *o1 = static_cast<TO>(*i1); \
    *o2 = static_cast<TO>(*i2); \
    *o3 = static_cast<TO>(*i3);

#define UNFOLD_ROUND_4()            \
    *o0 = half::round_to_half(*i0); \
    *o1 = half::round_to_half(*i1); \
    *o2 = half::round_to_half(*i2); \
    *o3 = half::round_to_half(*i3);

#define DIV_4(arg)                      \
    auto res = div((arg), 4);           \
    size_t res_quot = size_t(res.quot); \
    size_t res_rem = size_t(res.rem);

template <class TI, class TO>
inline void gnne_store_impl(const TI *input, TO *output, const runtime_shape_t &in_shape)
{
    auto size = detail::compute_size(in_shape);
    if constexpr (std::is_same_v<TI, TO>)
    {
        memcpy(output, input, size * sizeof(TI));
    }
    else if constexpr (std::is_same_v<TI, half> && std::is_same_v<TO, float>)
    {
        DIV_4(size)
        auto *output_ptr = output;
        const auto *input_ptr = input;
        for (size_t i = 0; i < res_quot; i++)
        {
            UNFOLD_ASSIGN_4(input_ptr, output_ptr);
            UNFOLD_CAST_4();
            input_ptr += 4;
            output_ptr += 4;
        }
        for (size_t i = 0; i < res_rem; i++)
        {
            *output_ptr = static_cast<float>(*input_ptr);
            output_ptr++;
            input_ptr++;
        }
    }
    else
    {
        std::runtime_error("Unsupported convert type in gnne_store!");
    }
}

#define GNNE_STORE_IMPL_LV2(input_t, output_t) \
    if (out_type == to_datatype<output_t>())   \
    gnne_store_impl(reinterpret_cast<const input_t *>(input), reinterpret_cast<output_t *>(output), in_shape)

#define GNNE_STORE_IMPL_LV1(input_t)                \
    if (in_type == to_datatype<input_t>())          \
    {                                               \
        GNNE_STORE_IMPL_LV2(input_t, uint8_t);      \
        else GNNE_STORE_IMPL_LV2(input_t, int8_t);  \
        else GNNE_STORE_IMPL_LV2(input_t, half);    \
        else GNNE_STORE_IMPL_LV2(input_t, float);   \
        else GNNE_STORE_IMPL_LV2(input_t, int16_t); \
    }

template <class TI, class TO>
inline void gnne_store([[maybe_unused]] datatype_t in_type, datatype_t out_type, const TI *input, TO *output, const runtime_shape_t &in_shape)
{
    GNNE_STORE_IMPL_LV1(uint8_t)
    else GNNE_STORE_IMPL_LV1(int8_t) else GNNE_STORE_IMPL_LV1(half) else GNNE_STORE_IMPL_LV1(float) else GNNE_STORE_IMPL_LV1(int16_t) else std::runtime_error("unsupported data type in gnne_store!");
}

template <class TI, class TO>
inline void gnne_load_impl(const TI *input, TO *output, const runtime_shape_t &in_shape)
{
    auto size = detail::compute_size(in_shape);
    if constexpr (std::is_same_v<TI, TO>)
    {
        memcpy(output, input, size * sizeof(TI));
    }
    else if constexpr (std::is_same_v<TI, float> && std::is_same_v<TO, half>)
    {
        DIV_4(size)
        auto *output_ptr = output;
        const auto *input_ptr = input;
        for (size_t i = 0; i < res_quot; i++)
        {
            UNFOLD_ASSIGN_4(input_ptr, output_ptr);
            UNFOLD_ROUND_4();
            input_ptr += 4;
            output_ptr += 4;
        }
        for (size_t i = 0; i < res_rem; i++)
        {
            *output_ptr = half::round_to_half(*input_ptr);
            output_ptr++;
            input_ptr++;
        }
    }
    else
    {
        std::runtime_error("Unsupported convert type in gnne_load!");
    }
}

#define GNNE_LOAD_IMPL_LV2(input_t, output_t) \
    if (out_type == to_datatype<output_t>())  \
    gnne_load_impl(reinterpret_cast<const input_t *>(input), reinterpret_cast<output_t *>(output), in_shape)

#define GNNE_LOAD_IMPL_LV1(input_t)                \
    if (in_type == to_datatype<input_t>())         \
    {                                              \
        GNNE_LOAD_IMPL_LV2(input_t, uint8_t);      \
        else GNNE_LOAD_IMPL_LV2(input_t, int8_t);  \
        else GNNE_LOAD_IMPL_LV2(input_t, half);    \
        else GNNE_LOAD_IMPL_LV2(input_t, float);   \
        else GNNE_LOAD_IMPL_LV2(input_t, int16_t); \
    }

template <class TI, class TO>
inline void gnne_load(datatype_t in_type, [[maybe_unused]] datatype_t out_type, const TI *input, TO *output, const runtime_shape_t &in_shape)
{
    GNNE_LOAD_IMPL_LV1(uint8_t)
    else GNNE_LOAD_IMPL_LV1(int8_t) else GNNE_LOAD_IMPL_LV1(half) else GNNE_LOAD_IMPL_LV1(float) else GNNE_LOAD_IMPL_LV1(int16_t) else std::runtime_error("unsupported data type in gnne_load!");
}

template <class T>
inline T _apply_gnne_activation(float value, int32_t shift_bits, T x0, T kl, T bl, T kr, T br)
{
    if (value < x0)
    {
        if constexpr (std::is_same_v<T, half>)
            return half::round_to_half(value * kl / (1LL << shift_bits) + bl);
        else if constexpr (std::is_same_v<T, float>)
            return value * kl / (1LL << shift_bits) + bl;
        else
            throw std::runtime_error("Invalid act parameters");
    }
    else
    {
        if constexpr (std::is_same_v<T, half>)
            return half::round_to_half(value * kr / (1LL << shift_bits) + br);
        else if constexpr (std::is_same_v<T, float>)
            return value * kr / (1LL << shift_bits) + br;
        else
            throw std::runtime_error("Invalid act parameters");
    }
}

template <class T1, class T2>
inline T1 apply_act0(T1 value, const T2 *act_data, size_t channel, int32_t shift_bits)
{
    auto begin = 7 * channel;
    auto v = _apply_gnne_activation(
        value,
        shift_bits,
        static_cast<float>(act_data[begin + 6]), // x0
        static_cast<float>(act_data[begin + 0]), // kl
        static_cast<float>(act_data[begin + 2]), // bl
        static_cast<float>(act_data[begin + 1]), // kr
        static_cast<float>(act_data[begin + 3])); // br
    return detail::apply_activation(v, { act_data[begin + 4], act_data[begin + 5] });
}

template <size_t N, class T1, class T2>
inline T1 apply_multi_segments_act1(T1 value, const T2 *act_data)
{
    // threshold..., k..., b...
    auto act_process = [&](size_t index) {
        auto k = act_data[(N - 1) + index];
        auto b = act_data[(N - 1) + N + index];
        return value * k + b;
    };

    for (size_t i = 0; i < N - 1; ++i)
    {
        if (value < act_data[i])
        {
            return act_process(i);
        }
    }
    return act_process(N - 1);
}

template <bool Is16Segments, class T1, class T2>
inline T1 apply_act1(T1 value, const T2 *act_data, [[maybe_unused]] size_t channel)
{
    T1 v;
    if (Is16Segments)
    {
        v = apply_multi_segments_act1<16>(value, act_data);
        auto offset = 16 * 2 + 15;
        return detail::apply_activation(v, { act_data[offset + 0], act_data[offset + 1] });
    }
    else
    {
        auto begin = 7 * channel;
        v = _apply_gnne_activation(
            value,
            0,
            static_cast<float>(act_data[begin + 0]), // x0
            static_cast<float>(act_data[begin + 1]), // kl
            static_cast<float>(act_data[begin + 3]), // bl
            static_cast<float>(act_data[begin + 2]), // kr
            static_cast<float>(act_data[begin + 4])); // br
        return detail::apply_activation(v, { act_data[begin + 5], act_data[begin + 6] });
    }
}

template <bool Is16Segments>
inline void fake_gnne_activation(const float *input_a, const float *input_b, bool has_uninitailized_input, float *output, const half *act, const runtime_shape_t &in_a_shape, const runtime_shape_t &in_b_shape, const runtime_shape_t &out_shape, int32_t out_channels, gnne_activation_type act_type)
{
    // one input
    if (has_uninitailized_input)
    {
        if (in_a_shape == out_shape)
        {
            const auto img_size = in_a_shape[2] * in_a_shape[3];
            const auto batch_size = in_a_shape[1] * img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * batch_size;
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * img_size;
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t i = 0; i < img_size; i++)
                    {
                        float old_data = static_cast<float>(*input_a_ptr);
                        *output_ptr = static_cast<float>(apply_act1<Is16Segments>(old_data, act, oc));
                        ++input_a_ptr;
                        ++output_ptr;
                    }
                }
            }
        }
        // resize
        else
        {
            const auto in_img_size = in_a_shape[2] * in_a_shape[3];
            const auto out_img_size = out_shape[2] * out_shape[3];
            const auto in_batch_size = in_a_shape[1] * in_img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * in_batch_size;
                auto *begin_output_ptr = output + batch * out_channels * out_img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * in_img_size;
                    auto *output_ptr = begin_output_ptr + oc * out_img_size;
                    for (int out_h = 0; out_h < out_shape[2]; out_h++)
                    {
                        for (int out_w = 0; out_w < out_shape[3]; out_w++)
                        {
                            float old_data = static_cast<float>(*(input_a_ptr + (int32_t)floor(out_h * in_a_shape[2] / out_shape[2]) * in_a_shape[3] + (int32_t)floor(out_w * in_a_shape[3] / out_shape[3])));
                            *(output_ptr + out_h * out_shape[3] + out_w) = static_cast<float>(apply_act1<Is16Segments>(old_data, act, oc));
                        }
                    }
                }
            }
        }
    }
    // two inputs
    else
    {
        if (in_a_shape == in_b_shape)
        {
            const auto img_size = in_a_shape[2] * in_a_shape[3];
            const auto batch_size = in_a_shape[1] * img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * batch_size;
                auto *begin_input_b_ptr = input_b + batch * batch_size;
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * img_size;
                    auto *input_b_ptr = begin_input_b_ptr + oc * img_size;
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t i = 0; i < img_size; i++)
                    {
                        float old_data = 0.f;
                        if (act_type == gnne_activation_type::mul)
                            old_data = static_cast<float>(*input_a_ptr) * static_cast<float>(*input_b_ptr);
                        else
                            old_data = static_cast<float>(*input_a_ptr) + static_cast<float>(*input_b_ptr);

                        *output_ptr = static_cast<float>(apply_act1<Is16Segments>(old_data, act, oc));
                        ++input_a_ptr;
                        ++input_b_ptr;
                        ++output_ptr;
                    }
                }
            }
        }
        // broadcast
        else
        {
            const auto img_size = out_shape[2] * out_shape[3];
            for (size_t batch = 0; batch < out_shape[0]; batch++)
            {
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t oh = 0; oh < out_shape[2]; oh++)
                    {
                        for (size_t ow = 0; ow < out_shape[3]; ow++)
                        {
                            int input_a_batch_off = in_a_shape[0] == 1 ? 0 : batch * in_a_shape[1] * in_a_shape[2] * in_a_shape[3];
                            int input_a_oc_off = in_a_shape[1] == 1 ? 0 : oc * in_a_shape[2] * in_a_shape[3];
                            int input_a_oh_off = in_a_shape[2] == 1 ? 0 : oh * in_a_shape[3];
                            int input_a_ow_off = in_a_shape[3] == 1 ? 0 : ow;
                            int input_b_batch_off = in_b_shape[0] == 1 ? 0 : batch * in_b_shape[1] * in_b_shape[2] * in_b_shape[3];
                            int input_b_oc_off = in_b_shape[1] == 1 ? 0 : oc * in_b_shape[2] * in_b_shape[3];
                            int input_b_oh_off = in_b_shape[2] == 1 ? 0 : oh * in_b_shape[3];
                            int input_b_ow_off = in_b_shape[3] == 1 ? 0 : ow;
                            float old_data = 0.f;
                            float input_a_data = static_cast<float>(*(input_a + input_a_batch_off + input_a_oc_off + input_a_oh_off + input_a_ow_off));
                            float input_b_data = static_cast<float>(*(input_b + input_b_batch_off + input_b_oc_off + input_b_oh_off + input_b_ow_off));

                            if (act_type == gnne_activation_type::mul)
                                old_data = input_a_data * input_b_data;
                            else
                                old_data = input_a_data + input_b_data;
                            *output_ptr = static_cast<float>(apply_act1<Is16Segments>(old_data, act, oc));
                            ++output_ptr;
                        }
                    }
                }
            }
        }
    }
}

template <bool Is16Segments, class TIA, class TIB, class TO>
inline void gnne_activation_impl(const TIA *input_a, const TIB *input_b, bool has_uninitailized_input, TO *output, const half *act, const runtime_shape_t &in_a_shape, const runtime_shape_t &in_b_shape, const runtime_shape_t &out_shape, int32_t out_channels, quant_param_t deq_a_param, quant_param_t deq_b_param, gnne_activation_type act_type)
{
    // one input
    if (has_uninitailized_input)
    {
        if (in_a_shape == out_shape)
        {
            const auto img_size = in_a_shape[2] * in_a_shape[3];
            const auto batch_size = in_a_shape[1] * img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * batch_size;
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * img_size;
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t i = 0; i < img_size; i++)
                    {
                        float old_data = static_cast<float>(*input_a_ptr);
                        old_data -= deq_a_param.zero_point;
                        old_data *= deq_a_param.scale;
                        if constexpr (std::is_same_v<TO, half>)
                            *output_ptr = half::round_to_half(apply_act1<Is16Segments>(old_data, act, oc));
                        else if constexpr (std::is_same_v<TO, uint8_t>)
                            *output_ptr = (uint8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        else if constexpr (std::is_same_v<TO, int8_t>)
                            *output_ptr = (int8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        else
                            *output_ptr = (int16_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        ++input_a_ptr;
                        ++output_ptr;
                    }
                }
            }
        }
        // resize
        else
        {
            const auto in_img_size = in_a_shape[2] * in_a_shape[3];
            const auto out_img_size = out_shape[2] * out_shape[3];
            const auto in_batch_size = in_a_shape[1] * in_img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * in_batch_size;
                auto *begin_output_ptr = output + batch * out_channels * out_img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * in_img_size;
                    auto *output_ptr = begin_output_ptr + oc * out_img_size;
                    for (int out_h = 0; out_h < out_shape[2]; out_h++)
                    {
                        for (int out_w = 0; out_w < out_shape[3]; out_w++)
                        {
                            float old_data = static_cast<float>(*(input_a_ptr + (int32_t)floor(out_h * in_a_shape[2] / out_shape[2]) * in_a_shape[3] + (int32_t)floor(out_w * in_a_shape[3] / out_shape[3])));
                            old_data -= deq_a_param.zero_point;
                            old_data *= deq_a_param.scale;
                            if constexpr (std::is_same_v<TO, half>)
                                *(output_ptr + out_h * out_shape[3] + out_w) = half::round_to_half(apply_act1<Is16Segments>(old_data, act, oc));
                            else if constexpr (std::is_same_v<TO, uint8_t>)
                                *(output_ptr + out_h * out_shape[3] + out_w) = (uint8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                            else if constexpr (std::is_same_v<TO, int8_t>)
                                *(output_ptr + out_h * out_shape[3] + out_w) = (int8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                            else
                                *(output_ptr + out_h * out_shape[3] + out_w) = (int16_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        }
                    }
                }
            }
        }
    }
    // two inputs
    else
    {
        if (in_a_shape == in_b_shape)
        {
            const auto img_size = in_a_shape[2] * in_a_shape[3];
            const auto batch_size = in_a_shape[1] * img_size;
            for (size_t batch = 0; batch < in_a_shape[0]; batch++)
            {
                auto *begin_input_a_ptr = input_a + batch * batch_size;
                auto *begin_input_b_ptr = input_b + batch * batch_size;
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *input_a_ptr = begin_input_a_ptr + oc * img_size;
                    auto *input_b_ptr = begin_input_b_ptr + oc * img_size;
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t i = 0; i < img_size; i++)
                    {
                        float old_data = 0.f;
                        if (act_type == gnne_activation_type::mul)
                            old_data = (static_cast<float>(*input_a_ptr) - deq_a_param.zero_point) * deq_a_param.scale * (static_cast<float>(*input_b_ptr) - deq_b_param.zero_point) * deq_b_param.scale;
                        else
                            old_data = (static_cast<float>(*input_a_ptr) - deq_a_param.zero_point) * deq_a_param.scale + (static_cast<float>(*input_b_ptr) - deq_b_param.zero_point) * deq_b_param.scale;

                        if constexpr (std::is_same_v<TO, half>)
                            *output_ptr = half::round_to_half(apply_act1<Is16Segments>(old_data, act, oc));
                        else if constexpr (std::is_same_v<TO, uint8_t>)
                            *output_ptr = (uint8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        else if constexpr (std::is_same_v<TO, int8_t>)
                            *output_ptr = (int8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        else
                            *output_ptr = (int16_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                        ++input_a_ptr;
                        ++input_b_ptr;
                        ++output_ptr;
                    }
                }
            }
        }
        // broadcast
        else
        {
            const auto img_size = out_shape[2] * out_shape[3];
            for (size_t batch = 0; batch < out_shape[0]; batch++)
            {
                auto *begin_output_ptr = output + batch * out_channels * img_size;
                for (int oc = 0; oc < out_channels; oc++)
                {
                    auto *output_ptr = begin_output_ptr + oc * img_size;
                    for (size_t oh = 0; oh < out_shape[2]; oh++)
                    {
                        for (size_t ow = 0; ow < out_shape[3]; ow++)
                        {
                            int input_a_batch_off = in_a_shape[0] == 1 ? 0 : batch * in_a_shape[1] * in_a_shape[2] * in_a_shape[3];
                            int input_a_oc_off = in_a_shape[1] == 1 ? 0 : oc * in_a_shape[2] * in_a_shape[3];
                            int input_a_oh_off = in_a_shape[2] == 1 ? 0 : oh * in_a_shape[3];
                            int input_a_ow_off = in_a_shape[3] == 1 ? 0 : ow;
                            int input_b_batch_off = in_b_shape[0] == 1 ? 0 : batch * in_b_shape[1] * in_b_shape[2] * in_b_shape[3];
                            int input_b_oc_off = in_b_shape[1] == 1 ? 0 : oc * in_b_shape[2] * in_b_shape[3];
                            int input_b_oh_off = in_b_shape[2] == 1 ? 0 : oh * in_b_shape[3];
                            int input_b_ow_off = in_b_shape[3] == 1 ? 0 : ow;
                            float old_data = 0.f;
                            float input_a_data = static_cast<float>(*(input_a + input_a_batch_off + input_a_oc_off + input_a_oh_off + input_a_ow_off));
                            float input_b_data = static_cast<float>(*(input_b + input_b_batch_off + input_b_oc_off + input_b_oh_off + input_b_ow_off));

                            if (act_type == gnne_activation_type::mul)
                                old_data = (input_a_data - deq_a_param.zero_point) * deq_a_param.scale * (input_b_data - deq_b_param.zero_point) * deq_b_param.scale;
                            else
                                old_data = (input_a_data - deq_a_param.zero_point) * deq_a_param.scale + (input_b_data - deq_b_param.zero_point) * deq_b_param.scale;

                            if constexpr (std::is_same_v<TO, half>)
                                *output_ptr = half::round_to_half(apply_act1<Is16Segments>(old_data, act, oc));
                            else if constexpr (std::is_same_v<TO, uint8_t>)
                                *output_ptr = (uint8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                            else if constexpr (std::is_same_v<TO, int8_t>)
                                *output_ptr = (int8_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                            else
                                *output_ptr = (int16_t)lrintf(apply_act1<Is16Segments>(old_data, act, oc));
                            ++output_ptr;
                        }
                    }
                }
            }
        }
    }
}

#define GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, output_t) \
    if (output_type == to_datatype<output_t>())                  \
        gnne_activation_impl<Is16Segments>(reinterpret_cast<const input_a_t *>(input_a), reinterpret_cast<const input_b_t *>(input_b), has_uninitailized_input, reinterpret_cast<output_t *>(output), act, in_a_shape, in_b_shape, out_shape, out_channels, deq_a_param, deq_b_param, act_type);

#define GNNE_ACTIVATION_IMPL_LV1(input_a_t, input_b_t)                                                                                                                                                                                                \
    if (input_a_type == to_datatype<input_a_t>() && input_b_type == to_datatype<input_b_t>())                                                                                                                                                         \
    {                                                                                                                                                                                                                                                 \
        GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, uint8_t)                                                                                                                                                                                       \
        else GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, int8_t) else GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, int16_t) else GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, half) else GNNE_ACTIVATION_IMPL_LV2(input_a_t, input_b_t, float) \
    }

template <bool Is16Segments, class TIA, class TIB, class TO>
inline void gnne_activation(const TIA *input_a, datatype_t input_a_type, const TIB *input_b, datatype_t input_b_type, bool has_uninitailized_input, TO *output, datatype_t output_type, const half *act,
    const runtime_shape_t &in_a_shape, const runtime_shape_t &in_b_shape, const runtime_shape_t &out_shape, int32_t out_channels, quant_param_t deq_a_param, quant_param_t deq_b_param, gnne_activation_type act_type)
{
    GNNE_ACTIVATION_IMPL_LV1(half, half)
    else GNNE_ACTIVATION_IMPL_LV1(half, uint8_t) else GNNE_ACTIVATION_IMPL_LV1(half, int8_t) else GNNE_ACTIVATION_IMPL_LV1(half, int16_t) else GNNE_ACTIVATION_IMPL_LV1(uint8_t, half) else GNNE_ACTIVATION_IMPL_LV1(uint8_t, uint8_t) else GNNE_ACTIVATION_IMPL_LV1(uint8_t, int8_t) else GNNE_ACTIVATION_IMPL_LV1(uint8_t, int16_t) else GNNE_ACTIVATION_IMPL_LV1(int8_t, half) else GNNE_ACTIVATION_IMPL_LV1(int8_t, uint8_t) else GNNE_ACTIVATION_IMPL_LV1(int8_t, int8_t) else GNNE_ACTIVATION_IMPL_LV1(int8_t, int16_t) else GNNE_ACTIVATION_IMPL_LV1(int16_t, half) else GNNE_ACTIVATION_IMPL_LV1(int16_t, uint8_t) else GNNE_ACTIVATION_IMPL_LV1(int16_t, int8_t) else GNNE_ACTIVATION_IMPL_LV1(int16_t, int16_t)
}

template <class T>
inline void gnne_transpose_impl(const T *input, T *output, const runtime_shape_t &in_shape, const mfu_trans_permute_t &gnne_perm)
{
    auto perm = to_axis(gnne_perm);
    runtime_shape_t out_shape(4);
    for (size_t i = 0; i < 4; i++)
    {
        out_shape[i] = in_shape[perm[i]];
    }
    DIV_4(out_shape[3])
    size_t move_distance = std::accumulate(in_shape.begin() + perm[3] + 1, in_shape.end(), 1, std::multiplies<size_t>());

    auto out_img_size = out_shape[2] * out_shape[3];
    for (size_t b = 0; b < out_shape[0]; b++)
    {
        // #pragma omp parallel for num_threads(kernels::default_kernel_context().num_threads)
        for (size_t c = 0; c < out_shape[1]; c++)
        {
            runtime_shape_t index(4);
            index[perm[0]] = b;
            index[perm[1]] = c;
            auto *output_ptr = output + b * out_shape[1] * out_img_size + c * out_img_size;
            for (size_t h = 0; h < out_shape[2]; h++)
            {
                index[perm[2]] = h;
                auto *input_ptr = input + linear_index(in_shape, index);
                for (size_t w = 0; w < res_quot; w++)
                {
                    auto *i0 = input_ptr + 0 * move_distance;
                    auto *i1 = input_ptr + 1 * move_distance;
                    auto *i2 = input_ptr + 2 * move_distance;
                    auto *i3 = input_ptr + 3 * move_distance;
                    UNFOLD_OUTPUT_ASSIGN_4(output_ptr);
                    *o0 = *i0;
                    *o1 = *i1;
                    *o2 = *i2;
                    *o3 = *i3;
                    input_ptr += 4 * move_distance;
                    output_ptr += 4;
                }
                for (size_t w = 0; w < res_rem; w++)
                {
                    *output_ptr = *input_ptr;
                    input_ptr += move_distance;
                    ++output_ptr;
                }
            }
        }
    }
}

template <class T>
inline void gnne_transpose(datatype_t input_type, const T *input, T *output, const runtime_shape_t &in_shape, const mfu_trans_permute_t &gnne_perm)
{
#define TRANSPOSE_IMPL(type)                                                                                           \
    gnne_transpose_impl(reinterpret_cast<const type *>(input), reinterpret_cast<type *>(output), in_shape, gnne_perm); \
    break;

    switch (runtime::get_bytes(input_type))
    {
    case 1:
        if (input_type == dt_uint8)
        {
            TRANSPOSE_IMPL(uint8_t);
        }
        else
        {
            TRANSPOSE_IMPL(int8_t);
        }
    case 2:
        if (input_type == dt_int16)
        {
            TRANSPOSE_IMPL(int16_t);
        }
        else if (input_type == dt_float16)
        {
            TRANSPOSE_IMPL(half);
        }
        else
        {
            throw std::invalid_argument("transpose type is not support");
        }
    default:
        std::invalid_argument("transpose type is not support");
    }
}

runtime_shape_t get_padded_shape(const runtime_shape_t &in_shape, const runtime_paddings_t &paddings)
{
    runtime_shape_t out_shape(in_shape.size());
    for (size_t i = 0; i < in_shape.size(); i++)
        out_shape[i] = (size_t)((int32_t)in_shape[i] + paddings[i].sum() + (in_shape[i] - 1) * paddings[i].interior);
    return out_shape;
}

runtime_shape_t get_in_index(const runtime_shape_t &index, const runtime_shape_t &in_shape, const runtime_paddings_t &paddings, pad_mode_t mode, bool &pad_element)
{
    runtime_shape_t in_index(index.size());
    pad_element = false;
    for (size_t i = 0; i < index.size(); i++)
    {
        auto &padding = paddings[i];
        if ((int32_t)index[i] < padding.before)
        {
            pad_element = true;
            if (mode == pad_reflect)
                in_index[i] = (size_t)padding.before - index[i];
            else if (mode == pad_symmetric)
                in_index[i] = (size_t)padding.before - index[i] - 1;
            else if (mode == pad_edge)
                in_index[i] = 0;
        }
        else
        {
            auto cnt_idx = (int32_t)index[i] - padding.before;
            if (cnt_idx > (int32_t)in_shape[i] - 1)
            {
                pad_element = true;
                if (mode == pad_reflect)
                    in_index[i] = in_shape[i] - 2 - ((size_t)cnt_idx - in_shape[i]);
                else if (mode == pad_symmetric)
                    in_index[i] = in_shape[i] - 1 - ((size_t)cnt_idx - in_shape[i]);
                else if (mode == pad_edge)
                    in_index[i] = in_shape[i] - 1;
            }
            else
            {
                in_index[i] = (size_t)cnt_idx;
            }
        }
    }

    return in_index;
}

template <class TShape, class Callable, class TIt>
void ai2d_pad_apply_impl(Callable &&callable, TShape index_prefix, TIt index_begin, TIt index_end) noexcept
{
    const auto head = *index_begin++;
    index_prefix.push_back(0);
    if (index_begin == index_end)
    {
        for (size_t i = 0; i < head; i++)
        {
            index_prefix.back() = i;
            callable(index_prefix);
        }
    }
    else
    {
        for (size_t i = 0; i < head; i++)
        {
            index_prefix.back() = i;
            ai2d_pad_apply_impl(std::forward<Callable>(callable), index_prefix, index_begin, index_end);
        }
    }
}

template <class TShape, class Callable>
void ai2d_pad_apply(const TShape &shape, Callable &&callable) noexcept
{
    ai2d_pad_apply_impl(std::forward<Callable>(callable), TShape(), shape.cbegin(), shape.cend());
}

inline void fake_ai2d_pad(const float *input, float *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides, const runtime_shape_t &out_strides, const runtime_paddings_t &paddings, pad_mode_t mode, float pad_value)
{
    auto out_shape = get_padded_shape(in_shape, paddings);
    ai2d_pad_apply(out_shape, [&](const runtime_shape_t &index) -> void {
        bool pad_element = false;
        auto in_index = get_in_index(index, in_shape, paddings, mode, pad_element);
        float value;
        if (!pad_element || mode != pad_constant)
            value = input[offset(in_strides, in_index)];
        else
            value = pad_value;
        output[offset(out_strides, index)] = value;
    });
}

template <class TI, class TO>
inline void ai2d_pad_impl(const TI *input, TO *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides,
    const runtime_shape_t &out_strides, const runtime_paddings_t &paddings, pad_mode_t mode, float pad_value, int32_t in_deq_bias, quant_param_t out_quant_param)
{
    auto out_shape = get_padded_shape(in_shape, paddings);
    ai2d_pad_apply(out_shape, [&](const runtime_shape_t &index) -> void {
        bool pad_element = false;
        auto in_index = get_in_index(index, in_shape, paddings, mode, pad_element);
        TO value;
        if (!pad_element || mode != pad_constant)
            value = (TO)((input[offset(in_strides, in_index)] - in_deq_bias) / out_quant_param.scale + out_quant_param.zero_point);
        else
            value = (TO)((pad_value - in_deq_bias) / out_quant_param.scale + out_quant_param.zero_point);
        output[offset(out_strides, index)] = value;
    });
}

template <class TI, class TO>
void ai2d_pad(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides, const runtime_shape_t &out_strides, const runtime_paddings_t &paddings, pad_mode_t mode, float pad_value, int32_t in_deq_bias, quant_param_t out_quant_param)
{
#define AI2D_PAD_IMPL(input_type, output_type) \
    ai2d_pad_impl(reinterpret_cast<const input_type *>(input), reinterpret_cast<output_type *>(output), in_shape, in_strides, out_strides, paddings, mode, pad_value, in_deq_bias, out_quant_param);

    if (input_type == dt_uint8 && output_type == dt_uint8)
        AI2D_PAD_IMPL(uint8_t, uint8_t)
    else if (input_type == dt_int8 && output_type == dt_int8)
        AI2D_PAD_IMPL(int8_t, int8_t)
    else if (input_type == dt_int16 && output_type == dt_int16)
        AI2D_PAD_IMPL(int16_t, int16_t)
    else if (input_type == dt_uint8 && output_type == dt_float16)
        AI2D_PAD_IMPL(uint8_t, half)
    else if (input_type == dt_int8 && output_type == dt_float16)
        AI2D_PAD_IMPL(int8_t, half)
    else if (input_type == dt_int16 && output_type == dt_float16)
        AI2D_PAD_IMPL(int16_t, half)
    else
        throw std::runtime_error("Invalid connector datatype.");
}

inline void fake_ai2d_resize_bilinear(const float *input, float *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides,
    const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, NNCASE_UNUSED bool half_pixel_centers)
{
    auto scales = kernels::detail::get_resize_scales(in_shape, out_h, out_w, align_corners);
    auto height_scale = scales.first;
    auto width_scale = scales.second;

    const float rounding_offset = std::numeric_limits<float>::is_integer ? .5f : .0f;
    runtime_shape_t in_index(4), out_index(4);

    auto get_input = [&](int32_t in_y, int32_t in_x) {
        in_index[2] = in_y;
        in_index[3] = in_x;
        return input[offset(in_strides, in_index)];
    };

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        in_index[0] = batch;
        out_index[0] = batch;
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            in_index[1] = oc;
            out_index[1] = oc;
            for (size_t oy = 0; oy < (size_t)out_h; oy++)
            {
                out_index[2] = oy;
                float in_y;
                int32_t in_y0, in_y1;
                kernels::detail::set_resize_bilinear(oy, height_scale, half_pixel_centers, in_shape[2], in_y, in_y0, in_y1);

                for (size_t ox = 0; ox < (size_t)out_w; ox++)
                {
                    out_index[3] = ox;
                    float in_x;
                    int32_t in_x0, in_x1;
                    kernels::detail::set_resize_bilinear(ox, width_scale, half_pixel_centers, in_shape[3], in_x, in_x0, in_x1);

                    auto v0 = get_input(in_y0, in_x0);
                    auto v1 = get_input(in_y1, in_x0);
                    auto v2 = get_input(in_y0, in_x1);
                    auto v3 = get_input(in_y1, in_x1);

                    auto a0 = (1 - (in_y - in_y0)) * (1 - (in_x - in_x0));
                    auto a1 = (in_y - in_y0) * (1 - (in_x - in_x0));
                    auto a2 = (1 - (in_y - in_y0)) * (in_x - in_x0);
                    auto a3 = (in_y - in_y0) * (in_x - in_x0);
                    output[offset(out_strides, out_index)] = (float)(v0 * a0 + v1 * a1 + v2 * a2 + v3 * a3 + rounding_offset);
                }
            }
        }
    }
}

inline void fake_ai2d_resize_nearest_neighbor(const float *input, float *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides,
    const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, bool half_pixel_centers)
{
    auto scales = kernels::detail::get_resize_scales(in_shape, out_h, out_w, align_corners);
    auto height_scale = scales.first;
    auto width_scale = scales.second;

    runtime_shape_t in_index(4), out_index(4);
    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        in_index[0] = batch;
        out_index[0] = batch;
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            in_index[1] = oc;
            out_index[1] = oc;
            for (size_t oy = 0; oy < (size_t)out_h; oy++)
            {
                auto in_y = kernels::detail::get_nearest_neighbor(oy, in_shape[2], height_scale, align_corners, half_pixel_centers);
                in_index[2] = in_y;
                out_index[2] = oy;

                for (size_t ox = 0; ox < (size_t)out_w; ox++)
                {
                    auto in_x = kernels::detail::get_nearest_neighbor(ox, in_shape[3], width_scale, align_corners, half_pixel_centers);
                    in_index[3] = in_x;
                    out_index[3] = ox;
                    output[offset(out_strides, out_index)] = input[offset(in_strides, in_index)];
                }
            }
        }
    }
}

template <class TI, class TO>
inline void ai2d_resize_bilinear_impl(const TI *input, TO *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides,
    const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, NNCASE_UNUSED bool half_pixel_centers, int32_t in_deq_bias, quant_param_t out_quant_param)
{
    auto scales = kernels::detail::get_resize_scales(in_shape, out_h, out_w, align_corners);
    auto height_scale = scales.first;
    auto width_scale = scales.second;

    const float rounding_offset = std::numeric_limits<float>::is_integer ? .5f : .0f;
    runtime_shape_t in_index(4), out_index(4);

    auto get_input = [&](int32_t in_y, int32_t in_x) {
        in_index[2] = in_y;
        in_index[3] = in_x;
        return input[offset(in_strides, in_index)];
    };

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        in_index[0] = batch;
        out_index[0] = batch;
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            in_index[1] = oc;
            out_index[1] = oc;
            for (size_t oy = 0; oy < (size_t)out_h; oy++)
            {
                out_index[2] = oy;
                float in_y;
                int32_t in_y0, in_y1;
                kernels::detail::set_resize_bilinear(oy, height_scale, half_pixel_centers, in_shape[2], in_y, in_y0, in_y1);

                for (size_t ox = 0; ox < (size_t)out_w; ox++)
                {
                    out_index[3] = ox;
                    float in_x;
                    int32_t in_x0, in_x1;
                    kernels::detail::set_resize_bilinear(ox, width_scale, half_pixel_centers, in_shape[3], in_x, in_x0, in_x1);

                    auto v0 = get_input(in_y0, in_x0) - in_deq_bias;
                    auto v1 = get_input(in_y1, in_x0) - in_deq_bias;
                    auto v2 = get_input(in_y0, in_x1) - in_deq_bias;
                    auto v3 = get_input(in_y1, in_x1) - in_deq_bias;

                    auto a0 = (1 - (in_y - in_y0)) * (1 - (in_x - in_x0));
                    auto a1 = (in_y - in_y0) * (1 - (in_x - in_x0));
                    auto a2 = (1 - (in_y - in_y0)) * (in_x - in_x0);
                    auto a3 = (in_y - in_y0) * (in_x - in_x0);
                    output[offset(out_strides, out_index)] = (TO)((TO)(v0 * a0 + v1 * a1 + v2 * a2 + v3 * a3 + rounding_offset) * out_quant_param.scale + out_quant_param.zero_point);
                }
            }
        }
    }
}

template <class TI, class TO>
inline void ai2d_resize_nearest_neighbor_impl(const TI *input, TO *output, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides,
    const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, NNCASE_UNUSED bool half_pixel_centers, int32_t in_deq_bias, quant_param_t out_quant_param)
{
    auto scales = kernels::detail::get_resize_scales(in_shape, out_h, out_w, align_corners);
    auto height_scale = scales.first;
    auto width_scale = scales.second;

    runtime_shape_t in_index(4), out_index(4);
    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        in_index[0] = batch;
        out_index[0] = batch;
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            in_index[1] = oc;
            out_index[1] = oc;
            for (size_t oy = 0; oy < (size_t)out_h; oy++)
            {
                auto in_y = kernels::detail::get_nearest_neighbor(oy, in_shape[2], height_scale, align_corners, half_pixel_centers);
                in_index[2] = in_y;
                out_index[2] = oy;

                for (size_t ox = 0; ox < (size_t)out_w; ox++)
                {
                    auto in_x = kernels::detail::get_nearest_neighbor(ox, in_shape[3], width_scale, align_corners, half_pixel_centers);
                    in_index[3] = in_x;
                    out_index[3] = ox;
                    output[offset(out_strides, out_index)] = (TO)((input[offset(in_strides, in_index)] - in_deq_bias) / out_quant_param.scale + out_quant_param.zero_point);
                }
            }
        }
    }
}

template <class TI, class TO>
void ai2d_resize_bilinear(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides, const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, bool half_pixel_centers, int32_t in_deq_bias, quant_param_t out_quant_param)
{
#define AI2D_RESIZE_BILINEAR_IMPL(input_type, output_type) \
    ai2d_resize_bilinear_impl(reinterpret_cast<const input_type *>(input), reinterpret_cast<output_type *>(output), in_shape, in_strides, out_strides, out_h, out_w, align_corners, half_pixel_centers, in_deq_bias, out_quant_param);

    if (input_type == dt_uint8 && output_type == dt_uint8)
        AI2D_RESIZE_BILINEAR_IMPL(uint8_t, uint8_t)
    else if (input_type == dt_int8 && output_type == dt_int8)
        AI2D_RESIZE_BILINEAR_IMPL(int8_t, int8_t)
    else if (input_type == dt_int16 && output_type == dt_int16)
        AI2D_RESIZE_BILINEAR_IMPL(int16_t, int16_t)
    else if (input_type == dt_uint8 && output_type == dt_float16)
        AI2D_RESIZE_BILINEAR_IMPL(uint8_t, half)
    else if (input_type == dt_int8 && output_type == dt_float16)
        AI2D_RESIZE_BILINEAR_IMPL(int8_t, half)
    else if (input_type == dt_int16 && output_type == dt_float16)
        AI2D_RESIZE_BILINEAR_IMPL(int16_t, half)
    else
        throw std::runtime_error("Invalid connector datatype.");
}

template <class TI, class TO>
void ai2d_resize_nearest_neighbor(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const runtime_shape_t &in_shape, const runtime_shape_t &in_strides, const runtime_shape_t &out_strides, int32_t out_h, int32_t out_w, bool align_corners, bool half_pixel_centers, int32_t in_deq_bias, quant_param_t out_quant_param)
{
#define AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(input_type, output_type) \
    ai2d_resize_nearest_neighbor_impl(reinterpret_cast<const input_type *>(input), reinterpret_cast<output_type *>(output), in_shape, in_strides, out_strides, out_h, out_w, align_corners, half_pixel_centers, in_deq_bias, out_quant_param);

    if (input_type == dt_uint8 && output_type == dt_uint8)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(uint8_t, uint8_t)
    else if (input_type == dt_int8 && output_type == dt_int8)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(int8_t, int8_t)
    else if (input_type == dt_int16 && output_type == dt_int16)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(int16_t, int16_t)
    else if (input_type == dt_uint8 && output_type == dt_float16)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(uint8_t, half)
    else if (input_type == dt_int8 && output_type == dt_float16)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(int8_t, half)
    else if (input_type == dt_int16 && output_type == dt_float16)
        AI2D_RESIZE_NEAREST_NEIGHBOR_IMPL(int16_t, half)
    else
        throw std::runtime_error("Invalid connector datatype.");
}

inline void fake_gnne_matmul_ref(const float *input_a, const float *input_b, float *output, const float *act, int32_t a_batches, int32_t a_rows, int32_t a_cols, int32_t b_batches, int32_t b_cols)
{
    if (a_batches == b_batches)
    {
        for (int32_t on = 0; on < a_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[on * a_rows * a_cols + oy * a_cols + i];
                        const auto b = input_b[on * a_cols * b_cols + i * b_cols + ox];
                        value = value + a * b;
                    }

                    output[on * a_rows * b_cols + oy * b_cols + ox] = apply_act0(value, act, on, 0);
                }
            }
        }
    }
    else if (a_batches > b_batches && b_batches == 1)
    {
        for (int32_t on = 0; on < a_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[on * a_rows * a_cols + oy * a_cols + i];
                        const auto b = input_b[i * b_cols + ox];
                        value = value + a * b;
                    }

                    output[on * a_rows * b_cols + oy * b_cols + ox] = apply_act0(value, act, on, 0);
                }
            }
        }
    }
    else if (a_batches < b_batches && a_batches == 1)
    {
        for (int32_t on = 0; on < b_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[oy * a_cols + i];
                        const auto b = input_b[on * a_cols * b_cols + i * b_cols + ox];
                        value = value + a * b;
                    }

                    output[on * a_rows * b_cols + oy * b_cols + ox] = apply_act0(value, act, on, 0);
                }
            }
        }
    }
    else
    {
        throw std::runtime_error("Invalid matmul");
    }
}

inline void fake_gnne_matmul(const float *input_a, const float *input_b, float *output, const float *act, int32_t a_batches, int32_t a_rows, int32_t a_cols, int32_t b_batches, int32_t b_cols)
{
    fake_gnne_matmul_ref(input_a, input_b, output, act, a_batches, a_rows, a_cols, b_batches, b_cols);
}

template <class TIA, class TIB, class TO>
inline void gnne_matmul_impl(const TIA *input_a, const TIB *input_b, TO *output, const half *act, const uint8_t *in_a_bias, int32_t a_batches, int32_t a_rows, int32_t a_cols, int32_t b_batches, int32_t b_cols, const std::vector<uint8_t> deq_b_bias, [[maybe_unused]] const int32_t in_a_shift_bits, [[maybe_unused]] const int32_t in_b_shift_bits)
{
    if (a_batches == b_batches)
    {
        for (int32_t on = 0; on < a_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[on * a_rows * a_cols + oy * a_cols + i];
                        const auto b = input_b[on * a_cols * b_cols + i * b_cols + ox];
                        value = value + ((float)a - (float)(*in_a_bias)) * ((float)b - (float)deq_b_bias[on]);
                    }

                    if constexpr (std::is_same_v<TO, half>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = half::round_to_half(apply_act0(value, act, on, in_b_shift_bits));
                    else if constexpr (std::is_same_v<TO, uint8_t> || std::is_same_v<TO, int8_t> || std::is_same_v<TO, int16_t>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = (TO)lrintf(apply_act0(value, act, on, in_b_shift_bits));
                }
            }
            in_a_bias += a_rows;
        }
    }
    else if (a_batches > b_batches && b_batches == 1)
    {
        for (int32_t on = 0; on < a_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[on * a_rows * a_cols + oy * a_cols + i];
                        const auto b = input_b[i * b_cols + ox];
                        value = value + ((float)a - (float)(*in_a_bias)) * ((float)b - (float)deq_b_bias[0]);
                    }

                    if constexpr (std::is_same_v<TO, half>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = half::round_to_half(apply_act0(value, act, on, in_b_shift_bits));
                    else if constexpr (std::is_same_v<TO, uint8_t> || std::is_same_v<TO, int8_t> || std::is_same_v<TO, int16_t>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = (TO)lrintf(apply_act0(value, act, on, in_b_shift_bits));
                }
            }
            in_a_bias += a_rows;
        }
    }
    else if (a_batches < b_batches && a_batches == 1)
    {
        for (int32_t on = 0; on < b_batches; on++)
        {
            for (int32_t oy = 0; oy < a_rows; oy++)
            {
                for (int32_t ox = 0; ox < b_cols; ox++)
                {
                    float value = 0.f;

                    for (int32_t i = 0; i < a_cols; i++)
                    {
                        const auto a = input_a[oy * a_cols + i];
                        const auto b = input_b[on * a_cols * b_cols + i * b_cols + ox];
                        value = value + ((float)a - (float)(*in_a_bias)) * ((float)b - (float)deq_b_bias[on]);
                    }

                    if constexpr (std::is_same_v<TO, half>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = half::round_to_half(apply_act0(value, act, on, in_b_shift_bits));
                    else if constexpr (std::is_same_v<TO, uint8_t> || std::is_same_v<TO, int8_t> || std::is_same_v<TO, int16_t>)
                        output[on * a_rows * b_cols + oy * b_cols + ox] = (TO)lrintf(apply_act0(value, act, on, in_b_shift_bits));
                }
            }
        }
    }
    else
    {
        throw std::runtime_error("Invalid matmul");
    }
}

#define GNNE_MATMUL_IMPL_LV2(input_a_t, input_b_t, output_t) \
    if (output_type == to_datatype<output_t>())              \
        gnne_matmul_impl(reinterpret_cast<const input_a_t *>(input_a), reinterpret_cast<const input_b_t *>(input_b), reinterpret_cast<output_t *>(output), act, in_a_bias, a_batches, a_rows, a_cols, b_batches, b_cols, deq_b_bias, in_a_shift_bits, in_b_shift_bits);

#define GNNE_MATMUL_IMPL_LV1(input_a_t, input_b_t)                                                                                                                             \
    if (input_a_type == to_datatype<input_a_t>() && input_b_type == to_datatype<input_b_t>())                                                                                  \
    {                                                                                                                                                                          \
        GNNE_MATMUL_IMPL_LV2(input_a_t, input_b_t, uint8_t)                                                                                                                    \
        else GNNE_MATMUL_IMPL_LV2(input_a_t, input_b_t, int8_t) else GNNE_MATMUL_IMPL_LV2(input_a_t, input_b_t, int16_t) else GNNE_MATMUL_IMPL_LV2(input_a_t, input_b_t, half) \
    }

template <class TIA, class TIB, class TO>
inline void gnne_matmul(const TIA *input_a, datatype_t input_a_type, TO *output, datatype_t output_type, const TIB *input_b, datatype_t input_b_type, const half *act, const uint8_t *in_a_bias,
    int32_t a_batches, int32_t a_rows, int32_t a_cols, int32_t b_batches, int32_t b_cols, const std::vector<uint8_t> deq_b_bias, const int32_t in_a_shift_bits, const int32_t in_b_shift_bits)
{
    GNNE_MATMUL_IMPL_LV1(uint8_t, uint8_t)
    else GNNE_MATMUL_IMPL_LV1(uint8_t, int8_t) else GNNE_MATMUL_IMPL_LV1(uint8_t, int16_t) else GNNE_MATMUL_IMPL_LV1(int8_t, uint8_t) else GNNE_MATMUL_IMPL_LV1(int8_t, int8_t) else GNNE_MATMUL_IMPL_LV1(int8_t, int16_t) else GNNE_MATMUL_IMPL_LV1(int16_t, uint8_t) else GNNE_MATMUL_IMPL_LV1(int16_t, int8_t)
}

inline void fake_gnne_conv2d_ref(const float *input, float *output, const float *weights, const float *act, const runtime_shape_t &in_shape,
    int32_t groups, int32_t out_channels, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, const scalar pad_value)
{
    const auto out_h = detail::get_windowed_output_size(in_shape[2], filter_h, stride_h, dilation_h, padding_h);
    const auto out_w = detail::get_windowed_output_size(in_shape[3], filter_w, stride_w, dilation_w, padding_w);
    const auto g_ic = in_shape[1] / groups;
    const auto g_oc = out_channels / groups;

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        const auto *in_batch_p = input + batch * in_shape[1] * in_shape[2] * in_shape[3];
        for (int32_t og = 0; og < groups; og++)
        {
            const auto *in_group_p = in_batch_p + og * g_ic * in_shape[2] * in_shape[3];
            const auto *w_group_p = weights + og * g_oc * g_ic * filter_h * filter_w;
            for (int32_t oc = 0; oc < g_oc; oc++)
            {
                const auto *w_oc_p = w_group_p + oc * g_ic * filter_h * filter_w;
                for (size_t oy = 0; oy < out_h; oy++)
                {
                    for (size_t ox = 0; ox < out_w; ox++)
                    {
                        const int32_t in_y_origin = (oy * stride_h) - padding_h.before;
                        const int32_t in_x_origin = (ox * stride_w) - padding_w.before;
                        const int32_t filter_y_start = std::max(0, (-in_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_y_end = std::min(filter_h, ((int32_t)in_shape[2] - in_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_x_start = std::max(0, (-in_x_origin + dilation_w - 1) / dilation_w);
                        const int32_t filter_x_end = std::min(filter_w, ((int32_t)in_shape[3] - in_x_origin + dilation_w - 1) / dilation_w);
                        float value = 0.f;

                        for (int32_t ic = 0; ic < (int32_t)g_ic; ic++)
                        {
                            const auto *in_c_p = in_group_p + ic * in_shape[2] * in_shape[3];
                            const auto *w_ic_p = w_oc_p + ic * filter_h * filter_w;
                            for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                            {
                                for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                                {
                                    const int32_t in_y = in_y_origin + dilation_h * ky;
                                    const int32_t in_x = in_x_origin + dilation_w * kx;
                                    float in_v = 0.f;
                                    if (in_x < 0 || in_x >= in_shape[3]
                                        || in_y < 0 || in_y >= in_shape[2])
                                        in_v = pad_value.as<float>();
                                    else
                                        in_v = in_c_p[in_y * in_shape[3] + in_x];

                                    const auto w = w_ic_p[ky * filter_w + kx];
                                    value = value + (float)in_v * w;
                                }
                            }
                        }

                        *output = apply_act0(value, act, og * g_oc + oc, 0);
                        output++;
                    }
                }
            }
        }
    }
}

inline void fake_gnne_conv2d(const float *input, float *output, const float *weights, const float *act, const runtime_shape_t &in_shape,
    int32_t groups, int32_t out_channels, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, const scalar pad_value)
{
    fake_gnne_conv2d_ref(input, output, weights, act, in_shape,
        groups, out_channels,
        filter_h, filter_w,
        stride_h, stride_w,
        dilation_h, dilation_w,
        padding_h, padding_w, pad_value);
}

template <class T1, class T2, class T3>
inline void gnne_conv2d_impl(const T1 *input, const T2 *weights, T3 *output, const half *act, const uint8_t *w_bias, const runtime_shape_t &in_shape, int32_t groups, int32_t out_channels,
    int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, const int32_t deq_bias, const int32_t shift_bits)
{
    const auto out_h = detail::get_windowed_output_size(in_shape[2], filter_h, stride_h, dilation_h, padding_h);
    const auto out_w = detail::get_windowed_output_size(in_shape[3], filter_w, stride_w, dilation_w, padding_w);
    const auto g_ic = in_shape[1] / groups;
    const auto g_oc = out_channels / groups;
    const uint8_t *w_bias_origin = w_bias;

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        w_bias = w_bias_origin;
        const auto *in_batch_p = input + batch * in_shape[1] * in_shape[2] * in_shape[3];
        for (int32_t og = 0; og < groups; og++)
        {
            const auto *in_group_p = in_batch_p + og * g_ic * in_shape[2] * in_shape[3];
            const auto *w_group_p = weights + og * g_oc * g_ic * filter_h * filter_w;
            for (int32_t oc = 0; oc < g_oc; oc++)
            {
                const auto *w_oc_p = w_group_p + oc * g_ic * filter_h * filter_w;
                for (size_t oy = 0; oy < out_h; oy++)
                {
                    for (size_t ox = 0; ox < out_w; ox++)
                    {
                        const int32_t in_y_origin = (oy * stride_h) - padding_h.before;
                        const int32_t in_x_origin = (ox * stride_w) - padding_w.before;
                        const int32_t filter_y_start = std::max(0, (-in_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_y_end = std::min(filter_h, ((int32_t)in_shape[2] - in_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_x_start = std::max(0, (-in_x_origin + dilation_w - 1) / dilation_w);
                        const int32_t filter_x_end = std::min(filter_w, ((int32_t)in_shape[3] - in_x_origin + dilation_w - 1) / dilation_w);
                        float value = 0.f;

                        for (int32_t ic = 0; ic < (int32_t)g_ic; ic++)
                        {
                            const auto *in_c_p = in_group_p + ic * in_shape[2] * in_shape[3];
                            const auto *w_ic_p = w_oc_p + ic * filter_h * filter_w;
                            for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                            {
                                for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                                {
                                    const int32_t in_y = in_y_origin + dilation_h * ky;
                                    const int32_t in_x = in_x_origin + dilation_w * kx;

                                    float in_v = 0.f;
                                    if (in_x < 0 || in_x >= in_shape[3]
                                        || in_y < 0 || in_y >= in_shape[2])
                                        in_v = pad_value.as<float>();
                                    else
                                        in_v = in_c_p[in_y * in_shape[3] + in_x];

                                    const auto w = w_ic_p[ky * filter_w + kx];
                                    value = value + ((float)in_v - deq_bias) * ((float)w - (float)(*w_bias));
                                }
                            }
                        }
                        if constexpr (std::is_same_v<T3, half>)
                            *output = half::round_to_half(apply_act0(value, act, og * g_oc + oc, shift_bits));
                        else if constexpr (std::is_same_v<T3, uint8_t> || std::is_same_v<T3, int8_t> || std::is_same_v<T3, int16_t>)
                            *output = (T3)lrintf(apply_act0(value, act, og * g_oc + oc, shift_bits));
                        output++;
                    }
                }
                w_bias++;
            }
        }
    }
}

#define GNNE_CONV2D_IMPL_LV2(input_t, w_t, output_t) \
    if (output_type == to_datatype<output_t>())      \
        gnne_conv2d_impl(reinterpret_cast<const input_t *>(input), reinterpret_cast<const w_t *>(weights), reinterpret_cast<output_t *>(output), act, w_bias, in_shape, groups, out_channels, filter_h, filter_w, stride_h, stride_w, dilation_h, dilation_w, padding_h, padding_w, pad_value, deq_bias, shift_bits);

#define GNNE_CONV2D_IMPL_LV1(input_t, w_t)                                                                                                             \
    if (input_type == to_datatype<input_t>() && w_type == to_datatype<w_t>())                                                                          \
    {                                                                                                                                                  \
        GNNE_CONV2D_IMPL_LV2(input_t, w_t, uint8_t)                                                                                                    \
        else GNNE_CONV2D_IMPL_LV2(input_t, w_t, int8_t) else GNNE_CONV2D_IMPL_LV2(input_t, w_t, int16_t) else GNNE_CONV2D_IMPL_LV2(input_t, w_t, half) \
    }

template <class TI, class TW, class TO>
inline void gnne_conv2d(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const TW *weights, datatype_t w_type, const half *act, const uint8_t *w_bias,
    const runtime_shape_t &in_shape, int32_t groups, int32_t out_channels, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, const scalar pad_value, const int32_t deq_bias, const int32_t shift_bits)
{
    GNNE_CONV2D_IMPL_LV1(uint8_t, uint8_t)
    else GNNE_CONV2D_IMPL_LV1(uint8_t, int8_t) else GNNE_CONV2D_IMPL_LV1(uint8_t, int16_t) else GNNE_CONV2D_IMPL_LV1(int8_t, uint8_t) else GNNE_CONV2D_IMPL_LV1(int8_t, int8_t) else GNNE_CONV2D_IMPL_LV1(int8_t, int16_t) else GNNE_CONV2D_IMPL_LV1(int16_t, uint8_t) else GNNE_CONV2D_IMPL_LV1(int16_t, int8_t)
}

inline void fake_gnne_conv2d_transpose_ref(const float *input, float *output, const float *weights, const float *act, const runtime_shape_t &in_shape,
    const int32_t groups, const runtime_shape_t &out_shape, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, [[maybe_unused]] const scalar pad_value)
{
    auto output_size = kernels::detail::compute_size(out_shape);
    auto out_tmp = std::make_unique<float[]>(output_size);
    std::fill(out_tmp.get(), out_tmp.get() + output_size, 0.f);

    const auto g_ic = in_shape[1] / groups;
    const auto g_oc = out_shape[1] / groups;

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        float *out_batch_p = out_tmp.get() + batch * out_shape[1] * out_shape[2] * out_shape[3];

        for (int32_t g = 0; g < groups; g++)
        {
            float *out_group_p = out_batch_p + g * g_oc * out_shape[2] * out_shape[3];
            const auto *w_group_p = weights + g * g_oc * g_ic * filter_h * filter_w;

            for (size_t ic = 0; ic < g_ic; ic++)
            {
                for (size_t iy = 0; iy < in_shape[2]; iy++)
                {
                    for (size_t ix = 0; ix < in_shape[3]; ix++)
                    {
                        const int32_t out_y_origin = (iy * stride_h) - padding_h.before;
                        const int32_t out_x_origin = (ix * stride_w) - padding_w.before;
                        const int32_t filter_y_start = std::max(0, (-out_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_y_end = std::min(filter_h, ((int32_t)out_shape[2] - out_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_x_start = std::max(0, (-out_x_origin + dilation_w - 1) / dilation_w);
                        const int32_t filter_x_end = std::min(filter_w, ((int32_t)out_shape[3] - out_x_origin + dilation_w - 1) / dilation_w);

                        float in_v = 0.f;
                        if (ix < 0 || ix >= in_shape[3]
                            || iy < 0 || iy >= in_shape[2])
                            in_v = pad_value.as<float>();
                        else
                            in_v = *input;
                        input++;

                        for (size_t oc = 0; oc < g_oc; oc++)
                        {
                            float *out_c_p = out_group_p + oc * out_shape[2] * out_shape[3];
                            const auto *w_oc_p = w_group_p + oc * g_ic * filter_h * filter_w;
                            const auto *w_ic_p = w_oc_p + ic * filter_h * filter_w;

                            for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                            {
                                for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                                {
                                    const int32_t out_y = out_y_origin + dilation_h * ky;
                                    const int32_t out_x = out_x_origin + dilation_w * kx;

                                    const auto w = w_ic_p[ky * filter_w + kx];

                                    out_c_p[out_y * out_shape[3] + out_x] += (float)in_v * w;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    auto channel_size = out_shape[2] * out_shape[3];
    for (size_t i = 0; i < output_size; i++)
    {
        auto oc = i / channel_size;
        output[i] = apply_act0(out_tmp[i], act, oc, 0);
    }
}

inline void fake_gnne_conv2d_transpose(const float *input, float *output, const float *weights, const float *act, const runtime_shape_t &in_shape,
    int32_t groups, const runtime_shape_t &out_shape, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, const scalar &pad_value)
{
    fake_gnne_conv2d_transpose_ref(input, output, weights, act, in_shape, groups, out_shape, filter_h, filter_w, stride_h, stride_w, dilation_h, dilation_w, padding_h, padding_w, pad_value);
}

template <class T1, class T2, class T3>
inline void gnne_conv2d_transpose_impl(const T1 *input, const T2 *weights, T3 *output, const half *act, const uint8_t *w_bias, const runtime_shape_t &in_shape, int32_t groups, const runtime_shape_t &out_shape,
    int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, const int32_t deq_bias, const int32_t shift_bits)
{
    auto output_size = kernels::detail::compute_size(out_shape);
    auto out_tmp = std::make_unique<float[]>(output_size);
    std::fill(out_tmp.get(), out_tmp.get() + output_size, 0.f);

    const auto g_ic = in_shape[1] / groups;
    const auto g_oc = out_shape[1] / groups;
    const uint8_t *w_bias_origin = w_bias;

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        w_bias = w_bias_origin;
        float *out_batch_p = out_tmp.get() + batch * out_shape[1] * out_shape[2] * out_shape[3];

        for (int32_t g = 0; g < groups; g++)
        {
            float *out_group_p = out_batch_p + g * g_oc * out_shape[2] * out_shape[3];
            const auto *w_group_p = weights + g * g_oc * g_ic * filter_h * filter_w;

            for (size_t ic = 0; ic < g_ic; ic++)
            {
                for (size_t iy = 0; iy < in_shape[2]; iy++)
                {
                    for (size_t ix = 0; ix < in_shape[3]; ix++)
                    {
                        const int32_t out_y_origin = (iy * stride_h) - padding_h.before;
                        const int32_t out_x_origin = (ix * stride_w) - padding_w.before;
                        const int32_t filter_y_start = std::max(0, (-out_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_y_end = std::min(filter_h, ((int32_t)out_shape[2] - out_y_origin + dilation_h - 1) / dilation_h);
                        const int32_t filter_x_start = std::max(0, (-out_x_origin + dilation_w - 1) / dilation_w);
                        const int32_t filter_x_end = std::min(filter_w, ((int32_t)out_shape[3] - out_x_origin + dilation_w - 1) / dilation_w);

                        float in_v = 0.f;
                        if (ix < 0 || ix >= in_shape[3]
                            || iy < 0 || iy >= in_shape[2])
                            in_v = pad_value.as<float>();
                        else
                            in_v = (float)*input;
                        input++;

                        for (size_t oc = 0; oc < g_oc; oc++)
                        {
                            float *out_c_p = out_group_p + oc * out_shape[2] * out_shape[3];
                            const auto *w_oc_p = w_group_p + oc * g_ic * filter_h * filter_w;
                            const auto *w_ic_p = w_oc_p + ic * filter_h * filter_w;

                            for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                            {
                                for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                                {
                                    const int32_t out_y = out_y_origin + dilation_h * ky;
                                    const int32_t out_x = out_x_origin + dilation_w * kx;

                                    const auto w = w_ic_p[ky * filter_w + kx];

                                    out_c_p[out_y * out_shape[3] + out_x] += ((float)in_v - deq_bias) * ((float)w - (float)(*w_bias));
                                }
                            }
                            w_bias++;
                        }
                    }
                }
            }
        }
    }

    auto channel_size = out_shape[2] * out_shape[3];
    for (size_t i = 0; i < output_size; i++)
    {
        auto oc = i / channel_size;
        if constexpr (std::is_same_v<T3, half>)
            output[i] = half::round_to_half(apply_act0(out_tmp[i], act, oc, shift_bits));
        else if constexpr (std::is_same_v<T3, uint8_t> || std::is_same_v<T3, int8_t> || std::is_same_v<T3, int16_t>)
            output[i] = (T3)lrintf(apply_act0(out_tmp[i], act, oc, shift_bits));
    }
}

#define GNNE_CONV2D_TRANSPOSE_IMPL_LV2(input_t, w_t, output_t) \
    if (output_type == to_datatype<output_t>())                \
        gnne_conv2d_transpose_impl(reinterpret_cast<const input_t *>(input), reinterpret_cast<const w_t *>(weights), reinterpret_cast<output_t *>(output), act, w_bias, in_shape, groups, out_shape, filter_h, filter_w, stride_h, stride_w, dilation_h, dilation_w, padding_h, padding_w, pad_value, deq_bias, shift_bits);

#define GNNE_CONV2D_TRANSPOSE_IMPL_LV1(input_t, w_t)                                                                                                                                 \
    if (input_type == to_datatype<input_t>() && w_type == to_datatype<w_t>())                                                                                                        \
    {                                                                                                                                                                                \
        GNNE_CONV2D_TRANSPOSE_IMPL_LV2(input_t, w_t, uint8_t)                                                                                                                        \
        else GNNE_CONV2D_TRANSPOSE_IMPL_LV2(input_t, w_t, int8_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV2(input_t, w_t, int16_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV2(input_t, w_t, half) \
    }

template <class TI, class TW, class TO>
inline void gnne_conv2d_transpose(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const TW *weights, datatype_t w_type, const half *act, const uint8_t *w_bias,
    const runtime_shape_t &in_shape, int32_t groups, const runtime_shape_t &out_shape, int32_t filter_h, int32_t filter_w, int32_t stride_h, int32_t stride_w, int32_t dilation_h, int32_t dilation_w,
    const padding &padding_h, const padding &padding_w, const scalar pad_value, const int32_t deq_bias, const int32_t shift_bits)
{
    GNNE_CONV2D_TRANSPOSE_IMPL_LV1(uint8_t, uint8_t)
    else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(uint8_t, int8_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(uint8_t, int16_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(int8_t, uint8_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(int8_t, int8_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(int8_t, int16_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(int16_t, uint8_t) else GNNE_CONV2D_TRANSPOSE_IMPL_LV1(int16_t, int8_t)
}

template <class T, class TBinaryOp, class TOutputOp>
void fake_pdp_reduce(const T *input, T *output, const runtime_shape_t &in_shape, int32_t filter_h, int32_t filter_w,
    int32_t stride_h, int32_t stride_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, bool count_include_pad, TBinaryOp &&binary_op, TOutputOp &&window_op)
{
    const auto out_h = kernels::detail::get_windowed_output_size(in_shape[2], filter_h, stride_h, 1, padding_h);
    const auto out_w = kernels::detail::get_windowed_output_size(in_shape[3], filter_w, stride_w, 1, padding_w);
    runtime_shape_t out_shape { in_shape[0], in_shape[1], (size_t)out_h, (size_t)out_w };

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            for (size_t oy = 0; oy < out_h; oy++)
            {
                for (size_t ox = 0; ox < out_w; ox++)
                {
                    const int32_t in_y_origin = (oy * stride_h) - padding_h.before;
                    const int32_t in_x_origin = (ox * stride_w) - padding_w.before;
                    const int32_t filter_y_start = std::max(0, -in_y_origin);
                    const int32_t filter_y_end = std::min(filter_h, (int32_t)in_shape[2] - in_y_origin);
                    const int32_t filter_x_start = std::max(0, -in_x_origin);
                    const int32_t filter_x_end = std::min(filter_w, (int32_t)in_shape[3] - in_x_origin);
                    auto value = input[linear_index(in_shape, { batch, oc, (size_t)(in_y_origin + filter_y_start), (size_t)((in_x_origin + filter_x_start)) })];
                    int32_t kernel_count = 0;

                    for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                    {
                        for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                        {
                            const int32_t in_y = in_y_origin + ky;
                            const int32_t in_x = in_x_origin + kx;

                            const auto in_v = input[linear_index(in_shape, { batch, oc, (size_t)in_y, (size_t)in_x })];

                            if (!((ky == filter_y_start) && (kx == filter_x_start)))
                                value = binary_op(value, in_v);
                            kernel_count++;
                        }
                    }

                    if (count_include_pad)
                    {
                        for (int32_t i = 0; i < filter_h * filter_w - kernel_count; i++)
                            value = binary_op(value, pad_value.as<T>());
                        kernel_count = filter_h * filter_w;
                    }

                    T kernel_count_v;
                    if constexpr (std::is_same_v<T, half>)
                    {
                        kernel_count_v = half::round_to_half(static_cast<float>(kernel_count));
                    }
                    else
                    {
                        kernel_count_v = static_cast<T>(kernel_count);
                    }
                    output[linear_index(out_shape, { batch, oc, (size_t)oy, (size_t)ox })] = window_op(value, kernel_count_v);
                }
            }
        }
    }
}

template <class TI, class TO, class TBinaryOp, class TOutputOp>
inline void gnne_pdp0_reduce_impl(const TI *input, half *act, TO *output, const runtime_shape_t &in_shape, int32_t filter_h, int32_t filter_w,
    int32_t stride_h, int32_t stride_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, dequantize_param_t deq_p, bool count_include_pad, int32_t shift_bits, TBinaryOp &&binary_op, TOutputOp &&window_op)
{
    const auto out_h = kernels::detail::get_windowed_output_size(in_shape[2], filter_h, stride_h, 1, padding_h);
    const auto out_w = kernels::detail::get_windowed_output_size(in_shape[3], filter_w, stride_w, 1, padding_w);
    runtime_shape_t out_shape { in_shape[0], in_shape[1], (size_t)out_h, (size_t)out_w };

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            for (size_t oy = 0; oy < out_h; oy++)
            {
                for (size_t ox = 0; ox < out_w; ox++)
                {
                    const int32_t in_y_origin = (oy * stride_h) - padding_h.before;
                    const int32_t in_x_origin = (ox * stride_w) - padding_w.before;
                    const int32_t filter_y_start = std::max(0, -in_y_origin);
                    const int32_t filter_y_end = std::min(filter_h, (int32_t)in_shape[2] - in_y_origin);
                    const int32_t filter_x_start = std::max(0, -in_x_origin);
                    const int32_t filter_x_end = std::min(filter_w, (int32_t)in_shape[3] - in_x_origin);
                    auto value = (float)(input[linear_index(in_shape, { batch, oc, (size_t)(in_y_origin + filter_y_start), (size_t)((in_x_origin + filter_x_start)) })]);
                    int32_t kernel_count = 0;

                    for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                    {
                        for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                        {
                            const int32_t in_y = in_y_origin + ky;
                            const int32_t in_x = in_x_origin + kx;

                            const auto in_v = input[linear_index(in_shape, { batch, oc, (size_t)in_y, (size_t)in_x })];

                            if (!((ky == filter_y_start) && (kx == filter_x_start)))
                                value = binary_op((float)(value - deq_p.zero_point), (float)(in_v - deq_p.zero_point));
                            kernel_count++;
                        }
                    }

                    if (count_include_pad)
                    {
                        for (int32_t i = 0; i < filter_h * filter_w - kernel_count; i++)
                            value = binary_op(value, pad_value.as<float>());
                        kernel_count = filter_h * filter_w;
                    }

                    TI kernel_count_v;
                    if constexpr (std::is_same_v<TI, half>)
                    {
                        kernel_count_v = half::round_to_half(static_cast<float>(kernel_count));
                    }
                    else
                    {
                        kernel_count_v = static_cast<TI>(kernel_count);
                    }
                    auto tmp_output = window_op(value * deq_p.scale, kernel_count_v);
                    tmp_output = apply_act0(tmp_output, act, oc, shift_bits);
                    if constexpr (std::is_same_v<TO, half>)
                        output[linear_index(out_shape, { batch, oc, (size_t)oy, (size_t)ox })] = half::round_to_half(tmp_output);
                    else
                        output[linear_index(out_shape, { batch, oc, (size_t)oy, (size_t)ox })] = (TO)std::lrintf(tmp_output);
                }
            }
        }
    }
}

template <class TI, class TO, class TBinaryOp, class TOutputOp>
void gnne_pdp0_reduce(const TI *input, half *act, TO *output, datatype_t input_type, datatype_t output_type, const runtime_shape_t &in_shape, int32_t filter_h, int32_t filter_w,
    int32_t stride_h, int32_t stride_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, dequantize_param_t deq_p, bool count_include_pad, int32_t shift_bits, TBinaryOp &&binary_op, TOutputOp &&window_op)
{
#define GNNE_PDP0_REDUCE_IMPL(input_type, output_type) \
    gnne_pdp0_reduce_impl(reinterpret_cast<const input_type *>(input), act, reinterpret_cast<output_type *>(output), in_shape, filter_h, filter_w, stride_h, stride_w, padding_h, padding_w, pad_value, deq_p, count_include_pad, shift_bits, binary_op, window_op);

    if (input_type == dt_uint8 && output_type == dt_uint8)
        GNNE_PDP0_REDUCE_IMPL(uint8_t, uint8_t)
    else if (input_type == dt_int8 && output_type == dt_int8)
        GNNE_PDP0_REDUCE_IMPL(int8_t, int8_t)
    else if (input_type == dt_uint8 && output_type == dt_float16)
        GNNE_PDP0_REDUCE_IMPL(uint8_t, half)
    else if (input_type == dt_int8 && output_type == dt_float16)
        GNNE_PDP0_REDUCE_IMPL(int8_t, half)
    else if (input_type == dt_int16 && output_type == dt_int16)
        GNNE_PDP0_REDUCE_IMPL(int16_t, int16_t)
    else if (input_type == dt_int16 && output_type == dt_float16)
        GNNE_PDP0_REDUCE_IMPL(int16_t, half)
    else
        throw std::runtime_error("Invalid connector datatype.");
}

template <class TI, class TO, class TBinaryOp, class TOutputOp>
inline void gnne_pdp1_impl(const TI *input, TO *output, const runtime_shape_t &in_shape, int32_t filter_h, int32_t filter_w,
    int32_t stride_h, int32_t stride_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, quant_param_t deq_p, quant_param_t q_p, bool count_include_pad, TBinaryOp &&binary_op, TOutputOp &&window_op)
{
    const auto out_h = kernels::detail::get_windowed_output_size(in_shape[2], filter_h, stride_h, 1, padding_h);
    const auto out_w = kernels::detail::get_windowed_output_size(in_shape[3], filter_w, stride_w, 1, padding_w);
    runtime_shape_t out_shape { in_shape[0], in_shape[1], (size_t)out_h, (size_t)out_w };

    for (size_t batch = 0; batch < in_shape[0]; batch++)
    {
        for (size_t oc = 0; oc < in_shape[1]; oc++)
        {
            for (size_t oy = 0; oy < out_h; oy++)
            {
                for (size_t ox = 0; ox < out_w; ox++)
                {
                    const int32_t in_y_origin = (oy * stride_h) - padding_h.before;
                    const int32_t in_x_origin = (ox * stride_w) - padding_w.before;
                    const int32_t filter_y_start = std::max(0, -in_y_origin);
                    const int32_t filter_y_end = std::min(filter_h, (int32_t)in_shape[2] - in_y_origin);
                    const int32_t filter_x_start = std::max(0, -in_x_origin);
                    const int32_t filter_x_end = std::min(filter_w, (int32_t)in_shape[3] - in_x_origin);
                    auto value = ((float)(input[linear_index(in_shape, { batch, oc, (size_t)(in_y_origin + filter_y_start), (size_t)((in_x_origin + filter_x_start)) })]) - deq_p.zero_point) * deq_p.scale;
                    int32_t kernel_count = 0;

                    for (int32_t ky = filter_y_start; ky < filter_y_end; ky++)
                    {
                        for (int32_t kx = filter_x_start; kx < filter_x_end; kx++)
                        {
                            const int32_t in_y = in_y_origin + ky;
                            const int32_t in_x = in_x_origin + kx;

                            const float in_v = ((float)(input[linear_index(in_shape, { batch, oc, (size_t)in_y, (size_t)in_x })]) - deq_p.zero_point) * deq_p.scale;

                            if (!((ky == filter_y_start) && (kx == filter_x_start)))
                                value = binary_op((float)value, (float)in_v);
                            kernel_count++;
                        }
                    }

                    if (count_include_pad)
                    {
                        for (int32_t i = 0; i < filter_h * filter_w - kernel_count; i++)
                            value = binary_op(value, pad_value.as<float>());
                        kernel_count = filter_h * filter_w;
                    }

                    half kernel_count_v = half::round_to_half(static_cast<float>(kernel_count));

                    auto tmp_output = window_op(value, kernel_count_v);
                    if constexpr (std::is_same_v<TO, half>)
                        output[linear_index(out_shape, { batch, oc, (size_t)oy, (size_t)ox })] = half::round_to_half(tmp_output);
                    else
                        output[linear_index(out_shape, { batch, oc, (size_t)oy, (size_t)ox })] = (TO)std::lrintf(half::round_to_half(tmp_output) * q_p.scale + q_p.zero_point);
                }
            }
        }
    }
}

template <class TI, class TO, class TBinaryOp, class TOutputOp>
void gnne_pdp1(const TI *input, datatype_t input_type, TO *output, datatype_t output_type, const runtime_shape_t &in_shape, int32_t filter_h, int32_t filter_w,
    int32_t stride_h, int32_t stride_w, const padding &padding_h, const padding &padding_w, const scalar pad_value, quant_param_t deq_p, quant_param_t q_p, bool count_include_pad, TBinaryOp &&binary_op, TOutputOp &&window_op)
{
#define GNNE_PDP1_IMPL(input_type, output_type) \
    gnne_pdp1_impl(reinterpret_cast<const input_type *>(input), reinterpret_cast<output_type *>(output), in_shape, filter_h, filter_w, stride_h, stride_w, padding_h, padding_w, pad_value, deq_p, q_p, count_include_pad, binary_op, window_op);

    if (input_type == dt_uint8 && output_type == dt_uint8)
        GNNE_PDP1_IMPL(uint8_t, uint8_t)
    else if (input_type == dt_uint8 && output_type == dt_int8)
        GNNE_PDP1_IMPL(uint8_t, int8_t)
    else if (input_type == dt_uint8 && output_type == dt_float16)
        GNNE_PDP1_IMPL(uint8_t, half)
    else if (input_type == dt_uint8 && output_type == dt_int16)
        GNNE_PDP1_IMPL(uint8_t, int16_t)
    else if (input_type == dt_int8 && output_type == dt_uint8)
        GNNE_PDP1_IMPL(int8_t, uint8_t)
    else if (input_type == dt_int8 && output_type == dt_int8)
        GNNE_PDP1_IMPL(int8_t, int8_t)
    else if (input_type == dt_int8 && output_type == dt_float16)
        GNNE_PDP1_IMPL(int8_t, half)
    else if (input_type == dt_int8 && output_type == dt_int16)
        GNNE_PDP1_IMPL(int8_t, int16_t)
    else if (input_type == dt_float16 && output_type == dt_uint8)
        GNNE_PDP1_IMPL(half, uint8_t)
    else if (input_type == dt_float16 && output_type == dt_int8)
        GNNE_PDP1_IMPL(half, int8_t)
    else if (input_type == dt_float16 && output_type == dt_float16)
        GNNE_PDP1_IMPL(half, half)
    else if (input_type == dt_float16 && output_type == dt_int16)
        GNNE_PDP1_IMPL(half, int16_t)
    else if (input_type == dt_int16 && output_type == dt_uint8)
        GNNE_PDP1_IMPL(int16_t, uint8_t)
    else if (input_type == dt_int16 && output_type == dt_int8)
        GNNE_PDP1_IMPL(int16_t, int8_t)
    else if (input_type == dt_int16 && output_type == dt_float16)
        GNNE_PDP1_IMPL(int16_t, half)
    else if (input_type == dt_int16 && output_type == dt_int16)
        GNNE_PDP1_IMPL(int16_t, int16_t)
    else
        throw std::runtime_error("Invalid connector datatype.");
}

template <class T>
inline void gnne_pad(const T *input, T *output, const runtime_shape_t &in_shape, const runtime_paddings_t &paddings, T pad_value, [[maybe_unused]] pad_mode_t mode)
{
    assert(mode == pad_constant);
    runtime_shape_t out_shape = { in_shape[0] + paddings[0].sum(),
        in_shape[1] + paddings[1].sum(),
        in_shape[2] + paddings[2].sum(),
        in_shape[3] + paddings[3].sum() };

    for (int d0 = 0; d0 < (int)out_shape[0]; d0++)
    {
        auto d0_origin = -paddings[0].before;
        auto in0 = input + ((int)d0_origin + d0) * in_shape[1] * in_shape[2] * in_shape[3];

        for (int d1 = 0; d1 < (int)out_shape[1]; d1++)
        {
            auto d1_origin = -paddings[1].before;
            auto in1 = in0 + ((int)d1_origin + d1) * in_shape[2] * in_shape[3];

            for (int d2 = 0; d2 < (int)out_shape[2]; d2++)
            {
                auto d2_origin = -paddings[2].before;
                auto in2 = in1 + ((int)d2_origin + d2) * in_shape[3];

                for (int d3 = 0; d3 < (int)out_shape[3]; d3++)
                {
                    auto d3_origin = -paddings[3].before;

                    if (d0 < paddings[0].before || d0 >= (int)out_shape[0] - paddings[0].after
                        || d1 < paddings[1].before || d1 >= (int)out_shape[1] - paddings[1].after
                        || d2 < paddings[2].before || d2 >= (int)out_shape[2] - paddings[2].after
                        || d3 < paddings[3].before || d3 >= (int)out_shape[3] - paddings[3].after)
                    {
                        *output++ = pad_value;
                    }
                    else
                        *output++ = in2[d3_origin + d3];
                }
            }
        }
    }
}

inline void fake_gnne_lstm(const float *input, const float *w_xc, const half *act_xc, const float *w_rc,
    const half *act_rc, const float *init_h, const float *init_c, const half *seg_fitting_param_ft,
    const half *seg_fitting_param_gt, float *output, float *output_h, float *output_c,
    const runtime_shape_t &in_shape, const runtime_shape_t &init_h_shape, const runtime_shape_t &init_c_shape,
    const runtime_shape_t &out_shape, const runtime_shape_t &w_xc_shape, const runtime_shape_t &w_rc_shape, lstm_direction direction)
{
    auto tanh = [&](float x) {
        for (size_t c = 0; c < 15; c++)
        {
            if (half(x) < seg_fitting_param_gt[c])
            {
                return float(seg_fitting_param_gt[15 + c] * x + seg_fitting_param_gt[15 + 16 + c]);
            }
        }
        return float(seg_fitting_param_gt[15 + 15] * x + seg_fitting_param_gt[15 + 16 + 15]);
    };
    auto sigmoid = [&](float x) {
        for (size_t c = 0; c < 15; c++)
        {
            if (half(x) < seg_fitting_param_ft[c])
            {
                return float(seg_fitting_param_ft[15 + c] * x + seg_fitting_param_ft[15 + 16 + c]);
            }
        }
        return float(seg_fitting_param_ft[15 + 15] * x + seg_fitting_param_ft[15 + 16 + 15]);
    };

    auto output_h_tmp = std::make_unique<float[]>(xt::compute_size(init_h_shape));
    auto output_c_tmp = std::make_unique<float[]>(xt::compute_size(init_c_shape));
    std::memcpy(output_h_tmp.get(), init_h, sizeof(float) * xt::compute_size(init_h_shape));
    std::memcpy(output_c_tmp.get(), init_c, sizeof(float) * xt::compute_size(init_c_shape));

    std::vector<uint32_t> seq_len_loop;
    for (uint32_t l = 0; l < in_shape[1]; l++)
        seq_len_loop.push_back(l);
    if (direction == lstm_direction::kReverse)
        std::reverse(seq_len_loop.begin(), seq_len_loop.end());
    for (uint32_t d = 0; d < out_shape[1]; d++)
    {
        if (d == 1)
            std::reverse(seq_len_loop.begin(), seq_len_loop.end());
        for (uint32_t b = 0; b < in_shape[2]; b++)
        {
            for (auto &l : seq_len_loop)
            {
                // g = w_xc_x + w_xc_x
                auto out_mul1 = std::vector<float>(out_shape[3] * 4);
                auto out_mul2 = std::vector<float>(out_shape[3] * 4);
                for (size_t o = 0; o < out_mul1.size(); o++)
                {
                    for (size_t i = 0; i < in_shape[3]; i++)
                    {
                        auto in_idx = i + b * in_shape[3] + l * in_shape[2] * in_shape[3];
                        auto w_idx = i + o * w_xc_shape[3] + d * w_xc_shape[2] * w_xc_shape[3];
                        out_mul1[o] += float(input[in_idx]) * float(w_xc[w_idx]);
                    }
                    auto base = d * w_xc_shape[2];
                    out_mul1[o] = apply_act0(out_mul1[o], act_xc, base + o, 0);

                    for (size_t i = 0; i < out_shape[3]; i++)
                    {
                        auto in_idx = i + b * out_shape[3] + d * out_shape[2] * out_shape[3];
                        auto w_idx = i + o * w_rc_shape[3] + d * w_rc_shape[2] * w_rc_shape[3];
                        out_mul2[o] += float(output_h_tmp[in_idx]) * float(w_rc[w_idx]);
                    }
                    base = d * w_rc_shape[2];
                    out_mul2[o] = apply_act0(out_mul2[o], act_rc, base + o, 0);

                    out_mul1[o] += out_mul2[o];
                }

                // ft = sigmoid(g[2])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 2] = sigmoid(out_mul1[o + out_shape[3] * 2]);
                }

                // ct = init_c * ft
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 2] = out_mul1[o + out_shape[3] * 2] * output_c_tmp[o + b * out_shape[3] + d * out_shape[2] * out_shape[3]];
                }

                // it = sigmoid(g[0])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 0] = sigmoid(out_mul1[o + out_shape[3] * 0]);
                }

                // c_t = tanh(g[3])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 3] = tanh(out_mul1[o + out_shape[3] * 3]);
                }

                // c_t_it = it * c_t
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 0] = out_mul1[o + out_shape[3] * 0] * out_mul1[o + out_shape[3] * 3];
                }

                // ct = ct + c_t_it
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    output_c_tmp[o + d * out_shape[2] * out_shape[3]] = out_mul1[o + out_shape[3] * 2] + out_mul1[o + out_shape[3] * 0];
                }

                // ot = sigmoid(g[1])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 1] = sigmoid(out_mul1[o + out_shape[3] * 1]);
                }

                // tanh_ct = tanh(ct_o)
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 3] = tanh(float(output_c_tmp[o + d * out_shape[2] * out_shape[3]]));
                }

                // ht = ot * tanh_ct
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    output_h_tmp[o + d * out_shape[2] * out_shape[3]] = out_mul1[o + out_shape[3] * 3] * out_mul1[o + out_shape[3] * 1];
                }
                std::memcpy(output + b * out_shape[3] + d * out_shape[2] * out_shape[3] + l * out_shape[1] * out_shape[2] * out_shape[3], output_h_tmp.get() + d * out_shape[2] * out_shape[3], sizeof(float) * out_shape[3]);

                if (l == seq_len_loop.back())
                {
                    std::memcpy(output_h + b * out_shape[3] + d * out_shape[2] * out_shape[3], output_h_tmp.get() + d * out_shape[2] * out_shape[3], sizeof(float) * out_shape[3]);
                    std::memcpy(output_c + b * out_shape[3] + d * out_shape[2] * out_shape[3], output_c_tmp.get() + d * out_shape[2] * out_shape[3], sizeof(float) * out_shape[3]);
                }
            }
        }
    }
}

template <class TI, class TIH, class TIC, class TW_XC, class TW_RC, class TO, class TOH, class TOC>
inline void gnne_lstm_impl(const TI *input, const TW_XC *w_xc, const half *act_xc,
    const TW_RC *w_rc, const half *act_rc_0, const half *act_rc_1, const TIH *init_h,
    const TIC *init_c, const half *seg_fitting_param_ft, const half *seg_fitting_param_gt,
    TO *output, TOH *output_h, TOC *output_c,
    const runtime_shape_t &in_shape, const runtime_shape_t &init_h_shape, const runtime_shape_t &init_c_shape,
    const runtime_shape_t &out_shape, const runtime_shape_t &w_xc_shape, const runtime_shape_t &w_rc_shape, lstm_direction direction,
    const half *w_xc_qarg, const half *w_rc_qarg, const half *bin_act, const half *bin_q_act,
    int32_t if_deq_bias, int32_t h_deq_bias_0, int32_t h_deq_bias_1, int32_t xc_shift_bits, int32_t rc_shift_bits_0, int32_t rc_shift_bits_1)
{
    auto tanh = [&](float x) {
        for (size_t c = 0; c < 15; c++)
        {
            if (half(x) < seg_fitting_param_gt[c])
            {
                return float(seg_fitting_param_gt[15 + c] * x + seg_fitting_param_gt[15 + 16 + c]);
            }
        }
        return float(seg_fitting_param_gt[15 + 15] * x + seg_fitting_param_gt[15 + 16 + 15]);
    };
    auto sigmoid = [&](float x) {
        for (size_t c = 0; c < 15; c++)
        {
            if (half(x) < seg_fitting_param_ft[c])
            {
                return float(seg_fitting_param_ft[15 + c] * x + seg_fitting_param_ft[15 + 16 + c]);
            }
        }
        return float(seg_fitting_param_ft[15 + 15] * x + seg_fitting_param_ft[15 + 16 + 15]);
    };

    auto output_h_tmp = std::make_unique<float[]>(xt::compute_size(init_h_shape));
    auto output_c_tmp = std::make_unique<float[]>(xt::compute_size(init_c_shape));
    std::memcpy(output_h_tmp.get(), init_h, sizeof(float) * xt::compute_size(init_h_shape));
    std::memcpy(output_c_tmp.get(), init_c, sizeof(float) * xt::compute_size(init_c_shape));

    std::vector<uint32_t> seq_len_loop;
    for (uint32_t l = 0; l < in_shape[1]; l++)
        seq_len_loop.push_back(l);
    if (direction == lstm_direction::kReverse)
        std::reverse(seq_len_loop.begin(), seq_len_loop.end());
    for (uint32_t d = 0; d < out_shape[1]; d++)
    {
        if (d == 1)
            std::reverse(seq_len_loop.begin(), seq_len_loop.end());
        for (uint32_t b = 0; b < in_shape[2]; b++)
        {
            for (auto &l : seq_len_loop)
            {
                // g = w_xc_x + w_xc_x
                auto out_mul1 = std::vector<float>(out_shape[3] * 4);
                auto out_mul2 = std::vector<float>(out_shape[3] * 4);
                for (size_t o = 0; o < out_mul1.size(); o++)
                {
                    for (size_t i = 0; i < in_shape[3]; i++)
                    {
                        auto in_idx = i + b * in_shape[3] + l * in_shape[2] * in_shape[3];
                        auto w_idx = i + o * w_xc_shape[3] + d * w_xc_shape[2] * w_xc_shape[3];
                        out_mul1[o] += float(input[in_idx] - if_deq_bias) * float(w_xc[w_idx] - w_xc_qarg[o]);
                    }
                    auto base = d * w_xc_shape[2];
                    out_mul1[o] = apply_act0(out_mul1[o] / (1 << xc_shift_bits), act_xc, base + o, 0);

                    for (size_t i = 0; i < out_shape[3]; i++)
                    {
                        auto in_idx = i + b * out_shape[3] + d * out_shape[2] * out_shape[3];
                        auto w_idx = i + o * w_rc_shape[3] + d * w_rc_shape[2] * w_rc_shape[3];
                        auto h_deq_bias = l == seq_len_loop.front() ? h_deq_bias_0 : h_deq_bias_1;
                        out_mul2[o] += float(output_h_tmp[in_idx] - h_deq_bias) * float(w_rc[w_idx] - w_rc_qarg[o]);
                    }
                    auto rc_shift_bits = l == seq_len_loop.front() ? rc_shift_bits_0 : rc_shift_bits_1;

                    base = d * w_rc_shape[2];
                    out_mul2[o] = apply_act0(out_mul2[o] / (1 << rc_shift_bits), l == seq_len_loop.front() ? act_rc_0 : act_rc_1, base + o, 0);

                    out_mul1[o] += out_mul2[o];
                }

                // ft = sigmoid(g[2])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 2] = sigmoid(out_mul1[o + out_shape[3] * 2]);
                }

                // ct = init_c * ft
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 2] = out_mul1[o + out_shape[3] * 2] * output_c_tmp[o + b * out_shape[3] + d * out_shape[2] * out_shape[3]];
                    out_mul1[o + out_shape[3] * 2] = apply_act1<false>(out_mul1[o + out_shape[3] * 2], bin_act, o);
                }

                // it = sigmoid(g[0])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 0] = sigmoid(out_mul1[o + out_shape[3] * 0]);
                }

                // c_t = tanh(g[3])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 3] = tanh(out_mul1[o + out_shape[3] * 3]);
                }

                // c_t_it = it * c_t
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 0] = out_mul1[o + out_shape[3] * 0] * out_mul1[o + out_shape[3] * 3];
                    out_mul1[o + out_shape[3] * 0] = apply_act1<false>(out_mul1[o + out_shape[3] * 0], bin_act, o);
                }

                // ct = ct + c_t_it
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    output_c_tmp[o + d * out_shape[2] * out_shape[3]] = out_mul1[o + out_shape[3] * 2] + out_mul1[o + out_shape[3] * 0];
                }

                // ot = sigmoid(g[1])
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 1] = sigmoid(out_mul1[o + out_shape[3] * 1]);
                }

                // tanh_ct = tanh(ct_o)
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    out_mul1[o + out_shape[3] * 3] = tanh(float(output_c_tmp[o]));
                }

                // ht = ot * tanh_ct
                for (size_t o = 0; o < out_shape[3]; o++)
                {
                    output_h_tmp[o + d * out_shape[2] * out_shape[3]] = out_mul1[o + out_shape[3] * 3] * out_mul1[o + out_shape[3] * 1];
                    output_h_tmp[o + d * out_shape[2] * out_shape[3]] = apply_act1<false>(output_h_tmp[o + d * out_shape[2] * out_shape[3]], bin_q_act, o);
                    auto output_h_int = int(std::round(output_h_tmp[o + d * out_shape[2] * out_shape[3]]));
                    if (std::is_same_v<TO, uint8_t>)
                        output[b * out_shape[3] + d * out_shape[2] * out_shape[3] + l * out_shape[1] * out_shape[2] * out_shape[3] + o] = TO(std::min(std::max(output_h_int, 0), 255));
                    else if (std::is_same_v<TO, int8_t>)
                        output[b * out_shape[3] + d * out_shape[2] * out_shape[3] + l * out_shape[1] * out_shape[2] * out_shape[3] + o] = TO(std::min(std::max(output_h_int, -127), 127));
                    else if (std::is_same_v<TO, int16_t>)
                        output[b * out_shape[3] + d * out_shape[2] * out_shape[3] + l * out_shape[1] * out_shape[2] * out_shape[3] + o] = TO(std::min(std::max(output_h_int, -2047), 2047));
                    if (l == seq_len_loop.back())
                        output_h[b * out_shape[3] + d * out_shape[2] * out_shape[3] + o] = TOH(output[b * out_shape[3] + d * out_shape[2] * out_shape[3] + l * out_shape[1] * out_shape[2] * out_shape[3] + o]);
                }

                if (l == seq_len_loop.back())
                {
                    for (size_t o = 0; o < out_shape[3]; o++)
                    {
                        output_c[b * out_shape[3] + d * out_shape[2] * out_shape[3] + o] = TOC(output_c_tmp[o]);
                    }
                }
            }
        }
    }
}

inline void gnne_lstm(const gsl::byte *input, datatype_t input_type, const gsl::byte *w_xc, datatype_t w_xc_type, const half *act_xc,
    const gsl::byte *w_rc, datatype_t w_rc_type, const half *act_rc_0, const half *act_rc_1, const gsl::byte *init_h, datatype_t init_h_type,
    const gsl::byte *init_c, datatype_t init_c_type, const half *seg_fitting_param_ft, const half *seg_fitting_param_gt,
    gsl::byte *output, datatype_t output_type, gsl::byte *output_h, datatype_t output_h_type, gsl::byte *output_c, datatype_t output_c_type,
    const runtime_shape_t &in_shape, const runtime_shape_t &init_h_shape, const runtime_shape_t &init_c_shape,
    const runtime_shape_t &out_shape, const runtime_shape_t &w_xc_shape, const runtime_shape_t &w_rc_shape, lstm_direction direction,
    const half *w_xc_qarg, const half *w_rc_qarg, const half *bin_act, const half *bin_q_act,
    int32_t if_deq_bias, int32_t h_deq_bias_0, int32_t h_deq_bias_1, int32_t xc_shift_bits, int32_t rc_shift_bits_0, int32_t rc_shift_bits_1)
{
#define GNNE_LSTM_IMPL(input_type, init_h_type, init_c_type, w_xc_type, w_rc_type, output_type, output_h_type, output_c_type) \
    gnne_lstm_impl(reinterpret_cast<const input_type *>(input), reinterpret_cast<const w_xc_type *>(w_xc), act_xc,            \
        reinterpret_cast<const w_rc_type *>(w_rc), act_rc_0, act_rc_1, reinterpret_cast<const init_h_type *>(init_h),         \
        reinterpret_cast<const init_c_type *>(init_c), seg_fitting_param_ft, seg_fitting_param_gt,                            \
        reinterpret_cast<output_type *>(output), reinterpret_cast<output_h_type *>(output_h),                                 \
        reinterpret_cast<output_c_type *>(output_c), in_shape, init_h_shape, init_c_shape,                                    \
        out_shape, w_xc_shape, w_rc_shape, direction, w_xc_qarg, w_rc_qarg, bin_act, bin_q_act, if_deq_bias, h_deq_bias_0,    \
        h_deq_bias_1, xc_shift_bits, rc_shift_bits_0, rc_shift_bits_1);

    if (input_type == dt_uint8 and w_xc_type == dt_uint8 and w_rc_type == w_xc_type and init_h_type == w_xc_type
        and output_type == w_xc_type and output_h_type == w_xc_type and init_c_type == dt_float16 and output_c_type == dt_float16)
    {
        GNNE_LSTM_IMPL(uint8_t, uint8_t, half, uint8_t, uint8_t, uint8_t, uint8_t, half)
    }
    else if (input_type == dt_int8 and w_xc_type == dt_int8 and w_rc_type == w_xc_type and init_h_type == w_xc_type
        and output_type == w_xc_type and output_h_type == w_xc_type and init_c_type == dt_float16 and output_c_type == dt_float16)
    {
        GNNE_LSTM_IMPL(int8_t, int8_t, half, int8_t, int8_t, int8_t, int8_t, half)
    }
    else if (input_type == dt_uint8 and w_xc_type == dt_int8 and w_rc_type == w_xc_type and init_h_type == w_xc_type
        and output_type == w_xc_type and output_h_type == w_xc_type and init_c_type == dt_float16 and output_c_type == dt_float16)
    {
        GNNE_LSTM_IMPL(uint8_t, int8_t, half, int8_t, int8_t, int8_t, int8_t, half)
    }
    else if (input_type == dt_int8 and w_xc_type == dt_uint8 and w_rc_type == w_xc_type and init_h_type == w_xc_type
        and output_type == w_xc_type and output_h_type == w_xc_type and init_c_type == dt_float16 and output_c_type == dt_float16)
    {
        GNNE_LSTM_IMPL(int8_t, uint8_t, half, uint8_t, uint8_t, uint8_t, uint8_t, half)
    }
    else
    {
        throw std::runtime_error("Invalid lstm datatype.");
    }
}

}